"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import copy
import pickle
import random
import unittest

import pytest

from flex.data.lazy_indexable import LazyIndexable

DEFAULT_LENGTH = 10
TEST_INIT_LEN_GUESS = 3
TEST_BIGGER_INIT_LEN_GUESS = 150


def get_generator(length=DEFAULT_LENGTH):
    return (x for x in range(length))


def get_iterator(length=DEFAULT_LENGTH):
    return iter(range(length))


def get_list(length=DEFAULT_LENGTH):
    return list(range(length))


def same_contents(reference, indexable: LazyIndexable):
    return all(x == y for x, y in zip(reference, indexable))


def same_contents_with_length(
    reference, indexable: LazyIndexable, length=DEFAULT_LENGTH
):
    return (
        all(x == y for x, y in zip(reference, indexable)) and len(indexable) == length
    )


def same_contects_negative_indexing(
    reference, indexable: LazyIndexable, length=DEFAULT_LENGTH
):
    return all(indexable[i - length] == x for i, x in enumerate(reference))


def same_contents_slicing(reference, indexable: LazyIndexable):
    return all(indexable[i:][0] == x for i, x in enumerate(reference))


def same_contents_anidated_slicing(reference, indexable: LazyIndexable):
    result = []
    anidated_slice = indexable
    for i, x in enumerate(reference):
        if i > 0 and i < (DEFAULT_LENGTH - 1):
            anidated_slice = anidated_slice[1:]
            result.append(anidated_slice[0] == x)
    return all(result)


def same_contects_split_in_two_slices(reference, indexable: LazyIndexable, split_index):
    result = []
    slice_1, slice_2 = indexable[:split_index], indexable[split_index:]
    for j, x in enumerate(reference):
        if j < split_index:
            result.append(slice_1[j] == x)
        else:
            result.append(slice_2[j - split_index] == x)
    return all(result)


class TestLazySliceable(unittest.TestCase):
    def test_copy_operator_from_array(self):
        base_list = get_list()
        from_array = LazyIndexable(base_list, length=DEFAULT_LENGTH)
        from_array_copy = copy.deepcopy(from_array)
        assert same_contents_with_length(from_array, from_array_copy)

    def test_copy_operator_from_iter(self):
        base_iter = get_iterator()
        from_iter = LazyIndexable(base_iter, length=DEFAULT_LENGTH)
        from_iter_copy = copy.deepcopy(from_iter)
        assert same_contents_with_length(from_iter, from_iter_copy)

    def test_copy_operator_from_generator(self):
        generator = get_generator()
        from_gen = LazyIndexable(generator, length=DEFAULT_LENGTH)
        from_gen_copy = copy.deepcopy(from_gen)
        assert same_contents_with_length(from_gen, from_gen_copy)

    def test_copy_operator_from_array_to_numpy(self):
        base_list = get_list()
        from_array = LazyIndexable(base_list, length=DEFAULT_LENGTH)
        from_array_copy = copy.deepcopy(from_array)
        assert same_contents_with_length(
            from_array.to_numpy(), from_array_copy.to_numpy()
        )

    def test_copy_operator_from_iter_to_numpy(self):
        base_iter = get_iterator()
        from_iter = LazyIndexable(base_iter, length=DEFAULT_LENGTH)
        from_iter_copy = copy.deepcopy(from_iter)
        assert same_contents_with_length(
            from_iter.to_numpy(), from_iter_copy.to_numpy()
        )

    def test_copy_operator_from_generator_to_numpy(self):
        generator = get_generator()
        from_gen = LazyIndexable(generator, length=DEFAULT_LENGTH)
        from_gen_copy = copy.deepcopy(from_gen)
        assert same_contents_with_length(from_gen.to_numpy(), from_gen_copy.to_numpy())

    def test_from_array(self):
        base_list = get_list()
        from_array = LazyIndexable(base_list, length=DEFAULT_LENGTH)
        assert same_contents(base_list, from_array)

    def test_from_generator(self):
        generator = get_generator()
        from_gen = LazyIndexable(get_generator(), length=DEFAULT_LENGTH)
        assert same_contents(generator, from_gen)
        from_gen = LazyIndexable(get_generator(), length=DEFAULT_LENGTH)
        assert same_contents(generator, from_gen)

    def test_from_iter(self):
        iterator = get_iterator()
        from_iter = LazyIndexable(get_iterator(), length=DEFAULT_LENGTH)
        assert same_contents(iterator, from_iter)
        from_iter = LazyIndexable(get_iterator(), length=DEFAULT_LENGTH)
        assert same_contents(iterator, from_iter)

    def test_negative_indexing_with_len_from_generator(self):
        generator = get_generator()
        from_gen = LazyIndexable(get_generator(), DEFAULT_LENGTH)
        assert same_contects_negative_indexing(generator, from_gen)

    def test_negative_indexing_with_len_from_iter(self):
        iterator = get_iterator()
        from_iter = LazyIndexable(get_iterator(), DEFAULT_LENGTH)
        assert same_contects_negative_indexing(iterator, from_iter)

    def test_from_array_to_numpy(self):
        base_list = get_list()
        from_array = LazyIndexable(base_list, length=DEFAULT_LENGTH)
        from_array = from_array.to_numpy()
        assert same_contents(base_list, from_array)
        base_list = get_list()
        from_array = LazyIndexable(base_list, length=DEFAULT_LENGTH)
        from_array = from_array.to_numpy()
        assert same_contents(base_list, from_array)

    def test_from_generator_to_numpy(self):
        generator = get_generator()
        from_gen = LazyIndexable(get_generator(), length=DEFAULT_LENGTH)
        from_gen = from_gen.to_numpy()
        assert same_contents(generator, from_gen)
        generator = get_generator()
        from_gen = LazyIndexable(get_generator(), length=DEFAULT_LENGTH)
        from_gen = from_gen.to_numpy()
        assert same_contents(generator, from_gen)

    def test_from_generator_with_len_to_numpy(self):
        generator = get_generator()
        from_gen = LazyIndexable(
            get_generator(),
            length=DEFAULT_LENGTH,
        )
        from_gen = from_gen.to_numpy()
        assert len(from_gen) == DEFAULT_LENGTH
        assert same_contents(generator, from_gen)
        generator = get_generator()
        from_gen = LazyIndexable(
            get_generator(),
            length=DEFAULT_LENGTH,
        )
        from_gen = from_gen.to_numpy()
        assert same_contents_with_length(generator, from_gen)

    def test_from_iter_to_numpy(self):
        iterator = get_iterator()
        from_iter = LazyIndexable(get_iterator(), length=DEFAULT_LENGTH)
        from_iter = from_iter.to_numpy()
        assert same_contents_with_length(iterator, from_iter)
        iterator = get_iterator()
        from_iter = LazyIndexable(get_iterator(), length=DEFAULT_LENGTH)
        from_iter = from_iter.to_numpy()
        assert same_contents_with_length(iterator, from_iter)

    def test_from_iter_with_len_to_numpy(self):
        iterator = get_iterator()
        from_iter = LazyIndexable(
            get_iterator(),
            length=DEFAULT_LENGTH,
        )
        from_iter = from_iter.to_numpy()
        assert same_contents_with_length(iterator, from_iter)
        iterator = get_iterator()
        from_iter = LazyIndexable(
            get_iterator(),
            length=DEFAULT_LENGTH,
        )
        from_iter = from_iter.to_numpy()
        assert same_contents_with_length(iterator, from_iter)

    def test_slicing_from_generator(self):
        generator = get_generator()
        from_gen = LazyIndexable(get_generator(), length=DEFAULT_LENGTH)
        assert same_contents_slicing(generator, from_gen)
        generator = get_generator()
        from_gen = LazyIndexable(get_generator(), length=DEFAULT_LENGTH)
        assert same_contents_slicing(generator, from_gen)

    def test_slicing_with_len_from_generator(self):
        generator = get_generator()
        from_gen = LazyIndexable(get_generator(), DEFAULT_LENGTH)
        assert same_contents_slicing(generator, from_gen)

    def test_slicing_with_len_from_iterator(self):
        iterator = get_iterator()
        from_iter = LazyIndexable(get_iterator(), DEFAULT_LENGTH)
        assert same_contents_slicing(iterator, from_iter)

    def test_slicing_from_iterator(self):
        iterator = get_iterator()
        from_iter = LazyIndexable(get_iterator(), length=DEFAULT_LENGTH)
        assert same_contents_slicing(iterator, from_iter)
        iterator = get_iterator()
        from_iter = LazyIndexable(get_iterator(), length=DEFAULT_LENGTH)
        assert same_contents_slicing(iterator, from_iter)

    def test_anidated_slices_with_len_from_generator(self):
        generator = get_generator()
        from_gen = LazyIndexable(get_generator(), DEFAULT_LENGTH)
        assert same_contents_anidated_slicing(generator, from_gen)

    def test_anidated_slices_from_generator(self):
        generator = get_generator()
        from_gen = LazyIndexable(get_generator(), length=DEFAULT_LENGTH)
        assert same_contents_anidated_slicing(generator, from_gen)

    def test_anidated_slices_with_len_from_iterator(self):
        iterator = get_iterator()
        from_iter = LazyIndexable(get_iterator(), DEFAULT_LENGTH)
        assert same_contents_anidated_slicing(iterator, from_iter)

    def test_anidated_slices_from_iterator(self):
        iterator = get_iterator()
        from_iter = LazyIndexable(get_iterator(), length=DEFAULT_LENGTH)
        assert same_contents_anidated_slicing(iterator, from_iter)

    def test_complex_slicing_with_len_from_generator(self):
        for i in range(DEFAULT_LENGTH):
            generator = get_generator()
            from_gen = LazyIndexable(get_generator(), DEFAULT_LENGTH)
            assert same_contects_split_in_two_slices(
                generator, from_gen, i
            ), f"Error at iteration: {i}"

    def test_complex_slicing_from_generator(self):
        for i in range(DEFAULT_LENGTH):
            generator = get_generator()
            from_gen = LazyIndexable(get_generator(), length=DEFAULT_LENGTH)
            assert same_contects_split_in_two_slices(
                generator, from_gen, i
            ), f"Error at iteration: {i}"

    def test_complex_slicing_with_len_from_iterator(self):
        for i in range(DEFAULT_LENGTH):
            gen_iter = LazyIndexable(get_iterator(), DEFAULT_LENGTH)
            iterator = get_iterator()
            assert same_contects_split_in_two_slices(
                iterator, gen_iter, i
            ), f"Error at iteration: {i}"

    def test_complex_slicing_from_iterator(self):
        for i in range(DEFAULT_LENGTH):
            gen_iter = LazyIndexable(get_iterator(), length=DEFAULT_LENGTH)
            iterator = get_iterator()
            assert same_contects_split_in_two_slices(
                iterator, gen_iter, i
            ), f"Error at iteration: {i}"

    def test_random_indexing_with_len_from_iterator(self):
        length = 100
        sample_size = 10
        random.seed(sample_size)
        available_indexes = get_list(length)
        for i in range(50):
            selected_indexes = random.sample(available_indexes, sample_size)
            from_iter = LazyIndexable(get_iterator(length=length), length)
            assert same_contents_with_length(
                selected_indexes, from_iter[selected_indexes], length=sample_size
            ), f"Error at iteration: {i}"

    def test_random_negative_indexing_with_len_from_iterator(self):
        length = 100
        sample_size = 10
        random.seed(sample_size)
        available_indexes = get_list(length)
        for i in range(50):
            selected_indexes = random.sample(available_indexes[1:], sample_size)
            negative_indexes = [-i for i in selected_indexes]
            positive_equivalent = [i + length for i in negative_indexes]
            from_iter = LazyIndexable(get_iterator(length=length), length)
            assert same_contents_with_length(
                positive_equivalent, from_iter[negative_indexes], length=sample_size
            ), f"Error at iteration: {i}"

    def test_random_indexing_from_iterator(self):
        length = 100
        sample_size = 10
        random.seed(sample_size)
        available_indexes = get_list(length)
        for i in range(50):
            selected_indexes = random.sample(available_indexes, sample_size)
            from_iter = LazyIndexable(
                get_iterator(length=length),
                length=length,
            )
            assert same_contents(
                selected_indexes, from_iter[selected_indexes]
            ), f"Error at iteration: {i}"
        for i in range(50):
            selected_indexes = random.sample(available_indexes, sample_size)
            from_iter = LazyIndexable(
                get_iterator(length=length),
                length=length,
            )
            assert same_contents(
                selected_indexes, from_iter[selected_indexes]
            ), f"Error at iteration: {i}"

    def test_random_indexing_with_len_from_generator(self):
        length = 100
        sample_size = 10
        random.seed(sample_size)
        available_indexes = get_list(length)
        for i in range(50):
            selected_indexes = random.sample(available_indexes, sample_size)
            from_gen = LazyIndexable(get_generator(length=length), length)
            assert same_contents_with_length(
                selected_indexes, from_gen[selected_indexes], length=sample_size
            ), f"Error at iteration: {i}"

    def test_random_negative_indexing_with_len_from_generator(self):
        length = 100
        sample_size = 10
        random.seed(sample_size)
        available_indexes = get_list(length)
        for j in range(50):
            selected_indexes = random.sample(available_indexes[1:], sample_size)
            negative_indexes = [-i for i in selected_indexes]
            positive_equivalent = [i + length for i in negative_indexes]
            from_gen = LazyIndexable(get_generator(length=length), length)
            assert same_contents_with_length(
                positive_equivalent, from_gen[negative_indexes], length=sample_size
            ), f"Error at iteration: {j}"

    def test_random_indexing_from_generator(self):
        length = 100
        sample_size = 10
        random.seed(sample_size)
        available_indexes = get_list(length)
        for i in range(50):
            selected_indexes = random.sample(available_indexes, sample_size)
            from_gen = LazyIndexable(
                get_generator(length=length),
                length=length,
            )
            assert same_contents(
                selected_indexes, from_gen[selected_indexes]
            ), f"Error at iteration: {i}"
        for i in range(50):
            selected_indexes = random.sample(available_indexes, sample_size)
            from_gen = LazyIndexable(
                get_generator(length=length),
                length=length,
            )
            assert same_contents(
                selected_indexes, from_gen[selected_indexes]
            ), f"Error at iteration: {i}"

    def test_random_negative_indexing_with_len_from_iterator_to_numpy(self):
        length = 100
        sample_size = 10
        random.seed(sample_size)
        available_indexes = get_list(length)
        for i in range(50):
            selected_indexes = random.sample(available_indexes[1:], sample_size)
            negative_indexes = [-i for i in selected_indexes]
            positive_equivalent = [i + length for i in negative_indexes]
            from_iter = LazyIndexable(get_iterator(length=length), length)
            sampled_indexes = from_iter[negative_indexes].to_numpy()
            assert same_contents_with_length(
                positive_equivalent, sampled_indexes, length=sample_size
            ), f"Error at iteration: {i}"

    def test_random_indexing_from_iterator_to_numpy(self):
        length = 100
        sample_size = 10
        random.seed(sample_size)
        available_indexes = get_list(length)
        for i in range(50):
            selected_indexes = random.sample(available_indexes, sample_size)
            from_iter = LazyIndexable(get_iterator(length=length), length=length)
            assert same_contents(
                selected_indexes, from_iter[selected_indexes].to_numpy()
            ), f"Error at iteration: {i}"
        for i in range(50):
            selected_indexes = random.sample(available_indexes, sample_size)
            from_iter = LazyIndexable(
                get_iterator(length=length),
                length=length,
            )
            assert same_contents(
                selected_indexes, from_iter[selected_indexes].to_numpy()
            ), f"Error at iteration: {i}"

    def test_random_indexing_with_len_from_generator_to_numpy(self):
        length = 1000
        sample_size = length // 2
        trials = 100
        random.seed(sample_size)
        available_indexes = get_list(length)
        for i in range(trials):
            selected_indexes = random.sample(available_indexes, sample_size)
            from_gen = LazyIndexable(get_generator(length=length), length)
            assert same_contents_with_length(
                selected_indexes,
                from_gen[selected_indexes].to_numpy(),
                length=sample_size,
            ), f"Error at iteration: {i}"

    def test_random_negative_indexing_with_len_from_generator_to_numpy(self):
        length = 1000
        sample_size = 10
        trials = 100
        random.seed(sample_size)
        available_indexes = get_list(length)
        for i in range(trials):
            selected_indexes = random.sample(available_indexes[1:], sample_size)
            negative_indexes = [-i for i in selected_indexes]
            positive_equivalent = [i + length for i in negative_indexes]
            from_gen = LazyIndexable(get_generator(length=length), length)
            assert same_contents_with_length(
                positive_equivalent,
                from_gen[negative_indexes].to_numpy(),
                length=sample_size,
            ), f"Error at iteration: {i}"

    def test_random_indexing_from_generator_to_numpy(self):
        length = 100
        sample_size = 10
        trials = 100
        random.seed(sample_size)
        available_indexes = get_list(length)
        for i in range(trials):
            selected_indexes = random.sample(available_indexes, sample_size)
            from_gen = LazyIndexable(get_generator(length=length), length=length)
            assert same_contents(
                selected_indexes, from_gen[selected_indexes].to_numpy()
            ), f"Error at iteration: {i}"
        for i in range(trials):
            selected_indexes = random.sample(available_indexes, sample_size)
            from_gen = LazyIndexable(
                get_generator(length=length),
                length=length,
            )
            assert same_contents(
                selected_indexes, from_gen[selected_indexes].to_numpy()
            ), f"Error at iteration: {i}"

    def test_empty_slice_with_len_from_generator(self):
        from_gen = LazyIndexable(get_generator(), DEFAULT_LENGTH)[:0]
        with pytest.raises(IndexError):
            from_gen[0]
        from_gen = LazyIndexable(get_generator(), DEFAULT_LENGTH)[
            5:2
        ]  # start greater than stop
        with pytest.raises(IndexError):
            from_gen[0]

    def test_empty_slice_from_generator(self):
        from_gen = LazyIndexable(get_generator(), length=DEFAULT_LENGTH)[:0]
        with pytest.raises(IndexError):
            from_gen[0]
        from_gen = LazyIndexable(get_generator(), length=DEFAULT_LENGTH)[
            5:2
        ]  # start greater than stop
        with pytest.raises(IndexError):
            from_gen[0]

    def test_empty_slice_with_len_from_iterator(self):
        from_iter = LazyIndexable(get_iterator(), DEFAULT_LENGTH)
        with pytest.raises(IndexError):
            from_iter[:0][0]
        from_iter = LazyIndexable(get_iterator(), DEFAULT_LENGTH)[
            5:2
        ]  # start greater than stop
        with pytest.raises(IndexError):
            from_iter[0]

    def test_empty_slice_from_iterator(self):
        from_iter = LazyIndexable(get_iterator(), length=DEFAULT_LENGTH)
        with pytest.raises(IndexError):
            from_iter[:0][0]
        from_iter = LazyIndexable(get_iterator(), length=DEFAULT_LENGTH)[
            5:2
        ]  # start greater than stop
        with pytest.raises(IndexError):
            from_iter[0]

    def test_index_out_of_bounds_with_len_from_generator(self):
        from_gen = LazyIndexable(get_generator(), DEFAULT_LENGTH)
        with pytest.raises(IndexError):
            from_gen[DEFAULT_LENGTH * DEFAULT_LENGTH]

    def test_index_out_of_bounds_from_generator(self):
        from_gen = LazyIndexable(get_generator(), length=DEFAULT_LENGTH)
        with pytest.raises(IndexError):
            from_gen[DEFAULT_LENGTH * DEFAULT_LENGTH]

    def test_index_out_of_bounds_with_len_from_iterator(self):
        from_iter = LazyIndexable(get_iterator(), DEFAULT_LENGTH)
        with pytest.raises(IndexError):
            from_iter[DEFAULT_LENGTH * DEFAULT_LENGTH]

    def test_pickle_from_array(self):
        base_list = get_list()
        from_array = LazyIndexable(base_list, length=DEFAULT_LENGTH)
        from_array_pickle_copy = pickle.loads(pickle.dumps(from_array))
        assert same_contents_with_length(from_array, from_array_pickle_copy)

    def test_pickle_from_iter(self):
        base_iter = get_iterator()
        from_iter = LazyIndexable(base_iter, length=DEFAULT_LENGTH)
        from_iter_pickle_copy = pickle.loads(pickle.dumps(from_iter))
        assert same_contents_with_length(from_iter, from_iter_pickle_copy)

    def test_pickle_from_generator(self):
        base_generator = get_generator()
        from_generator = LazyIndexable(base_generator, length=DEFAULT_LENGTH)
        from_generator_pickle_copy = pickle.loads(pickle.dumps(from_generator))
        assert same_contents_with_length(from_generator, from_generator_pickle_copy)



import tensor_pb2 as _tensor_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Error(_message.Message):
    __slots__ = ("reason",)
    REASON_FIELD_NUMBER: _ClassVar[int]
    reason: str
    def __init__(self, reason: _Optional[str] = ...) -> None: ...

class ClientMessage(_message.Message):
    __slots__ = ("handshake_res", "get_weights_res", "send_weights_res", "train_res", "error", "eval_res", "health_ins")
    class HandshakeRes(_message.Message):
        __slots__ = ("status",)
        STATUS_FIELD_NUMBER: _ClassVar[int]
        status: int
        def __init__(self, status: _Optional[int] = ...) -> None: ...
    class GetWeightsRes(_message.Message):
        __slots__ = ("weights",)
        WEIGHTS_FIELD_NUMBER: _ClassVar[int]
        weights: _containers.RepeatedCompositeFieldContainer[_tensor_pb2.Tensor]
        def __init__(self, weights: _Optional[_Iterable[_Union[_tensor_pb2.Tensor, _Mapping]]] = ...) -> None: ...
    class SendWeightsRes(_message.Message):
        __slots__ = ("status",)
        STATUS_FIELD_NUMBER: _ClassVar[int]
        status: int
        def __init__(self, status: _Optional[int] = ...) -> None: ...
    class TrainRes(_message.Message):
        __slots__ = ("metrics",)
        class MetricsEntry(_message.Message):
            __slots__ = ("key", "value")
            KEY_FIELD_NUMBER: _ClassVar[int]
            VALUE_FIELD_NUMBER: _ClassVar[int]
            key: str
            value: float
            def __init__(self, key: _Optional[str] = ..., value: _Optional[float] = ...) -> None: ...
        METRICS_FIELD_NUMBER: _ClassVar[int]
        metrics: _containers.ScalarMap[str, float]
        def __init__(self, metrics: _Optional[_Mapping[str, float]] = ...) -> None: ...
    class EvalRes(_message.Message):
        __slots__ = ("metrics",)
        class MetricsEntry(_message.Message):
            __slots__ = ("key", "value")
            KEY_FIELD_NUMBER: _ClassVar[int]
            VALUE_FIELD_NUMBER: _ClassVar[int]
            key: str
            value: float
            def __init__(self, key: _Optional[str] = ..., value: _Optional[float] = ...) -> None: ...
        METRICS_FIELD_NUMBER: _ClassVar[int]
        metrics: _containers.ScalarMap[str, float]
        def __init__(self, metrics: _Optional[_Mapping[str, float]] = ...) -> None: ...
    class HealthPing(_message.Message):
        __slots__ = ("status",)
        STATUS_FIELD_NUMBER: _ClassVar[int]
        status: int
        def __init__(self, status: _Optional[int] = ...) -> None: ...
    HANDSHAKE_RES_FIELD_NUMBER: _ClassVar[int]
    GET_WEIGHTS_RES_FIELD_NUMBER: _ClassVar[int]
    SEND_WEIGHTS_RES_FIELD_NUMBER: _ClassVar[int]
    TRAIN_RES_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    EVAL_RES_FIELD_NUMBER: _ClassVar[int]
    HEALTH_INS_FIELD_NUMBER: _ClassVar[int]
    handshake_res: ClientMessage.HandshakeRes
    get_weights_res: ClientMessage.GetWeightsRes
    send_weights_res: ClientMessage.SendWeightsRes
    train_res: ClientMessage.TrainRes
    error: Error
    eval_res: ClientMessage.EvalRes
    health_ins: ClientMessage.HealthPing
    def __init__(self, handshake_res: _Optional[_Union[ClientMessage.HandshakeRes, _Mapping]] = ..., get_weights_res: _Optional[_Union[ClientMessage.GetWeightsRes, _Mapping]] = ..., send_weights_res: _Optional[_Union[ClientMessage.SendWeightsRes, _Mapping]] = ..., train_res: _Optional[_Union[ClientMessage.TrainRes, _Mapping]] = ..., error: _Optional[_Union[Error, _Mapping]] = ..., eval_res: _Optional[_Union[ClientMessage.EvalRes, _Mapping]] = ..., health_ins: _Optional[_Union[ClientMessage.HealthPing, _Mapping]] = ...) -> None: ...

class ServerMessage(_message.Message):
    __slots__ = ("get_weights_ins", "send_weights_ins", "train_ins", "error", "eval_ins", "health_ins")
    class GetWeightsIns(_message.Message):
        __slots__ = ("status",)
        STATUS_FIELD_NUMBER: _ClassVar[int]
        status: int
        def __init__(self, status: _Optional[int] = ...) -> None: ...
    class SendWeightsIns(_message.Message):
        __slots__ = ("weights",)
        WEIGHTS_FIELD_NUMBER: _ClassVar[int]
        weights: _containers.RepeatedCompositeFieldContainer[_tensor_pb2.Tensor]
        def __init__(self, weights: _Optional[_Iterable[_Union[_tensor_pb2.Tensor, _Mapping]]] = ...) -> None: ...
    class TrainIns(_message.Message):
        __slots__ = ("status",)
        STATUS_FIELD_NUMBER: _ClassVar[int]
        status: int
        def __init__(self, status: _Optional[int] = ...) -> None: ...
    class EvalIns(_message.Message):
        __slots__ = ("status",)
        STATUS_FIELD_NUMBER: _ClassVar[int]
        status: int
        def __init__(self, status: _Optional[int] = ...) -> None: ...
    class HealthPing(_message.Message):
        __slots__ = ("status",)
        STATUS_FIELD_NUMBER: _ClassVar[int]
        status: int
        def __init__(self, status: _Optional[int] = ...) -> None: ...
    GET_WEIGHTS_INS_FIELD_NUMBER: _ClassVar[int]
    SEND_WEIGHTS_INS_FIELD_NUMBER: _ClassVar[int]
    TRAIN_INS_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    EVAL_INS_FIELD_NUMBER: _ClassVar[int]
    HEALTH_INS_FIELD_NUMBER: _ClassVar[int]
    get_weights_ins: ServerMessage.GetWeightsIns
    send_weights_ins: ServerMessage.SendWeightsIns
    train_ins: ServerMessage.TrainIns
    error: Error
    eval_ins: ServerMessage.EvalIns
    health_ins: ServerMessage.HealthPing
    def __init__(self, get_weights_ins: _Optional[_Union[ServerMessage.GetWeightsIns, _Mapping]] = ..., send_weights_ins: _Optional[_Union[ServerMessage.SendWeightsIns, _Mapping]] = ..., train_ins: _Optional[_Union[ServerMessage.TrainIns, _Mapping]] = ..., error: _Optional[_Union[Error, _Mapping]] = ..., eval_ins: _Optional[_Union[ServerMessage.EvalIns, _Mapping]] = ..., health_ins: _Optional[_Union[ServerMessage.HealthPing, _Mapping]] = ...) -> None: ...



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import unittest

import numpy as np
import pytest

from flex.data.dataset import Dataset


class TestFlexDataObject(unittest.TestCase):
    def test_validate_from_torchtext_dataset(self):
        from torchtext.datasets import AG_NEWS

        data = AG_NEWS(split="train")
        fcd = Dataset.from_torchtext_dataset(data)
        fcd.validate()

    def test_validate_from_torchtext_dataset_raise_error(self):
        data = np.random.rand(100).reshape([20, 5])
        with pytest.raises(ValueError):
            Dataset.from_torchtext_dataset(data)

    def test_validate_from_huggingface_dataset(self):
        from datasets import load_dataset

        data = load_dataset("ag_news", split="train")
        X_columns = "text"
        label_column = "label"
        fcd = Dataset.from_huggingface_datasets(
            data, X_columns=X_columns, label_column=label_column
        )
        fcd.validate()

    def test_validate_from_huggingface_dataset_error(self):
        data = np.random.rand(100).reshape([20, 5])
        X_columns = "text"
        label_column = "label"
        with pytest.raises(ValueError):
            Dataset.from_huggingface_datasets(data, X_columns, label_column)



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from typing import Iterable

from flex.actors.actors import FlexActors, FlexRole


def client_server_architecture(clients_ids: Iterable, server_id: str = "server"):
    """Method to create a client-server architeture from an Iterable of clients
    ids given, and a server id (optional).

    This method will assing to each id from the Iterable the client-role,
    and will create a new actor that will be the server-aggregator that will
    orchestrate the learning phase.

    Args:
    -----
        clients_ids (Iterable): List with the IDs for the clients
        server_id (str, optional): ID for the server actor. Defaults to None.

    Returns:
    --------
        FlexActors: The actors with their roles assigned.
    """
    actors = FlexActors()

    for client_id in clients_ids:
        actors[client_id] = FlexRole.client

    actors[server_id] = FlexRole.server_aggregator

    return actors


def p2p_architecture(nodes_ids: list):
    """Method to create a peer-to-peer (p2p) architecture from an Iterable of
    nodes given.

    This method will assing all roles (client-aggregator-server) to every id from
    the Iterable, so each participant in the learning phase can act as client,
    aggregator and server.

    Args:
    -----
        clients_ids (Iterable): Iterable with the clients ids

    Returns:
    --------
        FlexActors: Actors with their role assigned.
    """
    actors = FlexActors()

    for client_id in nodes_ids:
        actors[client_id] = FlexRole.server_aggregator_client

    return actors



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from flex import data
from flex import actors
from flex import model
from flex import pool
from flex import datasets

__version__ = "0.6.0"



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import logging
from concurrent import futures
from queue import Empty, Queue
from threading import Event
from typing import Dict, Iterator, List, Optional

import grpc
import numpy as np
from grpc import RpcContext

from flex.distributed.common import toNumpyArray, toTensorList
from flex.distributed.proto.transport_pb2 import ClientMessage, ServerMessage
from flex.distributed.proto.transport_pb2_grpc import (
    FlexibleServicer,
    add_FlexibleServicer_to_server,
)

logger = logging.getLogger(__name__)


class ClientProxy:
    """
    Abstraction for working with a client connection.
    Encapsulates the GRPC request iterator providing both methods for sending and pooling messages.

    Args:
    ----
        id (str): The ID of the client.
        request_iterator (Iterator[ClientMessage]): The iterator for receiving client messages.
        communication_queue (Queue): The queue for sending messages to the client.
        register_queue (Queue): The queue for registering the client.

    Attributes:
    ----------
        id (str): The ID of the client.
        request_iterator (Iterator[ClientMessage]): The iterator for receiving client messages.
        communication_queue (Queue): The queue for sending messages to the client.
        register_queue (Queue): The queue for registering the client.
    """

    def __init__(
        self,
        id: str,
        request_iterator: Iterator[ClientMessage],
        communication_queue: Queue,
        disconnected_event: Event,
    ):
        self.id = id
        self.request_iterator = request_iterator
        self.communication_queue = communication_queue
        self.disconnected_event = disconnected_event

    def __repr__(self):
        return f"ClientProxy({self.id})"

    def put_message(self, message: ServerMessage):
        """
        Sends a message to the client.

        Args:
        ----
            message (ServerMessage): The message to be sent to the client.
        """
        logger.info(f"Sending message to {self.id}. Message: {message}")
        self.communication_queue.put(message)

    def pool_messages(self):
        """
        Pools messages from the client.

        Returns
        -------
            ClientMessage: The next message received from the client.
        """
        try:
            response = next(self.request_iterator)
            if response.WhichOneof("msg") == "error":
                logger.error(
                    f"Client {self.id} sent an error message: {response.error}"
                )
                raise StopIteration("Client disconnected")
            return response
        except StopIteration:
            logger.info(f"Client {self.id} disconnected due to no more messages")
            self.communication_queue.put("Client disconnected")
            pass

    def disconnect(self):
        self.disconnected_event.set()

    def __del__(self):
        self.disconnected_event.set()


class ClientManager:
    """
    The ClientManager class manages the collection of ClientProxys for the connected clients.
    It provides methods for registering clients, deleting clients, broadcasting messages to clients,
    and pooling messages from clients.

    When the server starts to run, it spawns two threads. One thread runs the registration method
    of the ClientManager, which listens for new client registrations and adds them to the collection
    of clients. The other thread waits for termination methods and listens for new connections.

    Attributes
    ----------
        _register_queue (Queue): The queue used for registering new clients.
        _clients (Dict[str, ClientProxy]): The collection of ClientProxys for the connected clients.
    """

    def __init__(self, register_queue: Queue, register_stop_event: Event):
        self._register_queue = register_queue
        self._clients: Dict[str, ClientProxy] = {}
        self._register_stop_event = register_stop_event

    def __len__(self):
        return len(self._clients)

    def get_ids(self) -> List[str]:
        """
        Returns a list of client IDs currently connected to the server.

        Returns
        -------
            List[str]: A list of client IDs.
        """
        return list(self._clients.keys())

    def delete_client(self, client_id):
        """
        Deletes a client from the collection of clients.

        Args:
        ----
            client_id: The ID of the client to delete.
        """
        if client_id in self._clients:
            del self._clients[client_id]

    def run_registration(self):
        """
        The registration method that runs in a separate thread.
        Listens for new client registrations and adds them to the collection of clients.
        """
        i = 0
        while not self._register_stop_event.is_set():
            try:
                message = self._register_queue.get(timeout=0.1)
                request_iterator, communication_queue, finishing_event = message
                self._clients[str(i)] = ClientProxy(
                    str(i), request_iterator, communication_queue, finishing_event
                )
                logger.info(f"Client {i} registered")
            except Empty:
                pass

        for client in self._clients.values():
            client.disconnect()

    def broadcast(self, message: ServerMessage, node_ids: Optional[List[str]] = None):
        """
        Broadcasts a message to all or a subset of clients.

        Args:
        ----
            message (ServerMessage): The message to broadcast.
            node_ids (Optional[List[str]]): The IDs of the clients to broadcast the message to.
                If None, the message will be broadcasted to all clients.
        """
        if node_ids:
            if not set(node_ids).issubset(self._clients):
                logger.warning("Some node IDs are not registered")

        for id, client in self._clients.items():
            if node_ids is None or id in node_ids:
                client.put_message(message)

    def pool_clients(self, node_ids: Optional[List[str]] = None):
        """
        Pools messages from all or a subset of clients.

        Args:
        ----
            node_ids (Optional[List[str]]): The IDs of the clients to pool messages from.
                If None, messages will be pooled from all clients.

        Returns:
        -------
            List[Tuple[Optional[ClientMessage], str]]: A list of tuples containing the pooled messages
            and their corresponding client IDs.
        """
        if node_ids:
            messages = [
                (client.pool_messages(), id)
                for id, client in self._clients.items()
                if id in node_ids
            ]
        else:
            messages = [
                (client.pool_messages(), id) for id, client in self._clients.items()
            ]

        messages_not_none = [(m, id) for m, id in messages if m is not None]
        for id in [id for m, id in messages if m is None]:
            logger.info(f"Unregistering client {id}")
            self.delete_client(id)

        return messages_not_none


class ServerServicer(FlexibleServicer):
    def __init__(self, queue: Queue):
        super().__init__()
        self._registration_queue = queue

    @staticmethod
    def _handshake(message: ClientMessage):
        if (
            message.WhichOneof("msg") != "handshake_res"
            or message.handshake_res.status != 200
        ):
            raise grpc.RpcError("No handshake message")

    def Send(
        self, request_iterator: Iterator[ClientMessage], context: RpcContext
    ) -> Iterator[ServerMessage]:
        first_request = next(request_iterator)
        self._handshake(first_request)
        communication_queue = Queue()
        finishing_event = Event()
        self._registration_queue.put(
            (request_iterator, communication_queue, finishing_event)
        )
        logger.info("Client connected")

        while finishing_event.is_set() is False:
            try:
                value = communication_queue.get(timeout=0.1)
                if value == "Client disconnected":
                    context.cancel()
                    break
                yield value
            except Empty:
                continue
            except Exception:
                context.cancel()
                break


class Server:
    """
    Server for distributed FLEXible environment.

    This class represents a server in a distributed FLEXible environment. It provides methods for managing clients,
    collecting weights, sending weights, training models, and evaluating models.
    """

    def __init__(self):
        self._registration_queue = Queue()
        self._stop_event = Event()
        self._servicer = ServerServicer(queue=self._registration_queue)
        self._manager = ClientManager(
            register_queue=self._registration_queue,
            register_stop_event=self._stop_event,
        )
        self._server = None

    def __len__(self):
        """
        Returns the number of registered clients.

        Returns
        -------
            int: The number of registered clients.
        """
        return len(self._manager)

    def get_ids(self) -> List[str]:
        """
        Returns a list of client IDs.

        Returns
        -------
            List[str]: A list of client IDs.
        """
        return self._manager.get_ids()

    def ping(self, node_ids: Optional[List[str]] = None):
        """
        Sends a ping message to the specified nodes and retrieves their health status.

        Args:
        ----
            node_ids (Optional[List[str]]): A list of node IDs to send the ping message to. If None, the ping message will be sent to all nodes.

        Returns:
        -------
            None
        """
        self._manager.broadcast(
            ServerMessage(health_ins=ServerMessage.HealthPing(status=200)),
            node_ids=node_ids,
        )
        messages = self._manager.pool_clients(node_ids=node_ids)
        assert all(m.WhichOneof("msg") == "health_ins" for m, _ in messages)

    def collect_weights(self, node_ids: Optional[List[str]] = None):
        """
        Collects weights from clients.

        Args:
        ----
            node_ids (Optional[List[str]]): Optional list of client IDs. If provided, only the specified clients will be
                used for weight collection.

        Returns:
        -------
            List[np.ndarray]: A list of collected weights.
        """
        self._manager.broadcast(
            ServerMessage(get_weights_ins=ServerMessage.GetWeightsIns(status=200)),
            node_ids=node_ids,
        )
        messages = self._manager.pool_clients(node_ids=node_ids)
        assert all(m.WhichOneof("msg") == "get_weights_res" for m, _ in messages)
        rv = []
        for message, _ in messages:
            weights = toNumpyArray(message.get_weights_res.weights)
            rv.append(weights)

        return rv

    def send_weights(
        self, weights: List[np.ndarray], node_ids: Optional[List[str]] = None
    ):
        """
        Sends weights to clients.

        Args:
        ----
            weights (List[np.ndarray]): A list of weights to send to clients.
            node_ids (Optional[List[str]]): Optional list of client IDs. If provided, only the specified clients will receive
                the weights.
        """
        self._manager.broadcast(
            ServerMessage(
                send_weights_ins=ServerMessage.SendWeightsIns(
                    weights=toTensorList([np.array(w) for w in weights])
                )
            ),
            node_ids=node_ids,
        )

        messages = self._manager.pool_clients(node_ids=node_ids)
        assert all(m.WhichOneof("msg") == "send_weights_res" for m, _ in messages)

    def train(
        self, node_ids: Optional[List[str]] = None
    ) -> Dict[str, Dict[str, float]]:
        """
        Trains models on clients.

        Args:
        ----
            node_ids (Optional[List[str]]): Optional list of client IDs. If provided, only the specified clients will be used
                for training.

        Returns:
        -------
            Dict[str, Dict[str, float]]: A dictionary mapping client IDs to training metrics.
        """
        self._manager.broadcast(
            ServerMessage(train_ins=ServerMessage.TrainIns(status=200)),
            node_ids=node_ids,
        )
        messages = self._manager.pool_clients(node_ids=node_ids)
        assert all(m.WhichOneof("msg") == "train_res" for m, _ in messages)
        return {id: m.train_res.metrics for m, id in messages}

    def eval(self, node_ids: Optional[List[str]] = None) -> Dict[str, Dict[str, float]]:
        """
        Evaluates models on clients.

        Args:
        ----
            node_ids (Optional[List[str]]): Optional list of client IDs. If provided, only the specified clients will be used
                for evaluation.

        Returns:
        -------
            Dict[str, Dict[str, float]]: A dictionary mapping client IDs to evaluation metrics.
        """
        self._manager.broadcast(
            ServerMessage(eval_ins=ServerMessage.EvalIns(status=200)),
            node_ids=node_ids,
        )
        messages = self._manager.pool_clients(node_ids=node_ids)
        assert all(m.WhichOneof("msg") == "eval_res" for m, _ in messages)
        return {id: m.eval_res.metrics for m, id in messages}

    def run(
        self,
        address: str = "[::]",
        port: int = 50051,
        ssl_private_key: str = None,
        ssl_certificate_chain: str = None,
        ssl_root_certificate: str = None,
        require_client_auth: bool = False,
    ):
        """
        Starts the server. Does not block the main thread.

        Note: This method may be called only once.

        Args:
        ----
            address (str): The address to bind the server to. Defaults to "[::]".
            port (int): The port to bind the server to. Defaults to 50051.
            ssl_private_key (str): The path to the SSL private key file. If provided along with
                `ssl_certificate_chain`, the server will use secure gRPC communication.
            ssl_certificate_chain (str): The path to the SSL certificate chain file. If provided along with
                `ssl_private_key`, the server will use secure gRPC communication.
            ssl_root_certificate (str): The path to the SSL root certificate file. Required if `require_client_auth`
                is set to True.
            require_client_auth (bool): Whether to require client authentication. Defaults to False.
        """
        if self._server is not None:
            raise RuntimeError("Server is already running")

        address_port = f"{address}:{port}"
        self._executor = futures.ThreadPoolExecutor(max_workers=10)
        self._server = grpc.server(self._executor)
        add_FlexibleServicer_to_server(self._servicer, self._server)

        if ssl_private_key is not None and ssl_certificate_chain is not None:
            if require_client_auth:
                assert (
                    ssl_root_certificate is not None
                ), "Root certificate must be provided if client authentication is required"
            self._server.add_secure_port(
                address=address_port,
                server_credentials=grpc.ssl_server_credentials(
                    [(ssl_private_key, ssl_certificate_chain)],
                    root_certificates=ssl_root_certificate,
                    require_client_auth=require_client_auth,
                ),
            )
        else:
            self._server.add_insecure_port(address=address_port)

        self._server.start()
        logger.info(f"Server started at {address_port}")
        # Start the registration and termination threads
        self._registration = self._executor.submit(self._manager.run_registration)
        self._termination = self._executor.submit(self._server.wait_for_termination)

    def stop(self):
        """
        Stops the server.
        """
        self._stop_event.set()
        if self._server is not None:
            event = self._server.stop(None)
            self._server = None
            event.wait()
        self._termination.cancel()
        self._registration.cancel()
        self._executor.shutdown(cancel_futures=True, wait=True)

    def __del__(self):
        self.stop()



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
"""File that contains some utils functions to test help us testing the datasets that works with FLEXible.
"""
import gc  # noqa: E402

from flex.data import Dataset, FedDataDistribution, FedDatasetConfig  # noqa: E402

config = FedDatasetConfig(
    seed=0,
    n_nodes=2,
    replacement=False,
    node_ids=["client_0", "client_1"],
)


def iterate_module_functions(module):
    """Function to get the functions that load a torchtext dataset.

    Args:
        module (module): torchtext.datasets

    Returns:
        List: A list that contains the name and the function of the
        datasets available in torchtext.datasets.
    """
    return [[name, val] for name, val in module.__dict__.items() if callable(val)]


def check_if_can_load_torchtext_dataset(list_datasets):
    """Function that recieve a list of torchtext datasets and check whether
    or not can be loaded to FLEXible with the methods available. For those
    that gives error, we keep the error.

    Args:
        list_datasets (list): List of list (name, func) containing the datasets
        that will be tested.

    Returns:
        list, list: The first list contains the names of each database that
        can be loaded into FLEXible. The second list contains the names of
        each database that gives error while trying to load it to FLEXible.

    Raises:
        - Gives error if the database can't be loaded to FLEXible.
        - Gives error if the database can't be loaded as: func(split='train')
    """
    valid_datasets = []
    wrong_datasets = []
    for name, func in list_datasets:
        print(f"Testing dataset: {name}")
        try:
            data = func(split="train")
            fld = Dataset.from_torchtext_dataset(data)
            fld.validate()
            flex_dataset = FedDataDistribution.from_pytorch_text_dataset(data, config)
            del flex_dataset
            flex_dataset = FedDataDistribution.from_config(fld, config)
            del flex_dataset
            valid_datasets.append([name, func, "-"])
            del data
            del fld
        except Exception as e:
            wrong_datasets.append([name, func, e])
    return valid_datasets, wrong_datasets


def check_if_can_load_hf_dataset():
    """Function that takes a list of huggingface datasets and check whether
    or not can be loaded to FLEXible with the methods available. For those
    that gives error, we keep the error. Also, the user may indicate a list of
    columns that will be the features of the model (X_column) and the columns
    that will be the label (y_column) of the model.

    To check the datasets, refer to PluggableDatasetsHuggingFace

    Returns:
        list, list: The first list contains the names of each database that
        can be loaded into FLEXible. The second list contains the names of
        each database that gives error while trying to load it to FLEXible.

    Raises:
        - Gives error if the database can't be loaded to FLEXible.
        - Gives error if the database can't be loaded as: func(split='train')
    """
    from datasets import load_dataset

    from flex.data.pluggable_datasets import PluggableHuggingFace

    valid_datasets = []
    wrong_datasets = []
    for dataset in PluggableHuggingFace:
        name, X_columns, y_column = dataset.value
        print(f"Testing dataset: {name}")
        try:
            data = load_dataset(name, split="train")
            flex_dataset = FedDataDistribution.from_config_with_huggingface_dataset(
                data, config, X_columns, y_column
            )
            del flex_dataset
            fld = Dataset.from_huggingface_dataset(data, X_columns, y_column)
            fld.validate()
            flex_dataset = FedDataDistribution.from_config(fld, config)
            del flex_dataset
            del data
            del fld
            valid_datasets.append([name, "-"])
        except Exception as e:
            wrong_datasets.append([name, e])
    return valid_datasets, wrong_datasets


def check_if_can_load_text_tfds():
    """Function that take a list of tensorflow datasets and check whether
    or not can be loaded to FLEXible with the methods available. For those
    that gives error, we keep the error. Also, the user may indicate a list of
    columns that will be the features of the model (X_column) and the columns
    that will be the label (y_column) of the model.

    To check the datasets refer to PluggableDatasetsTensorFlowText

    Returns:
        list, list: The first list contains the names of each database that
        can be loaded into FLEXible. The second list contains the names of
        each database that gives error while trying to load it to FLEXible.

    Raises:
        - Gives error if the database can't be loaded to FLEXible.
        - Gives error with Question Answering datasets. Probably because
        they needs further preprocessing than other tasks.
            -> Example of an error:  AttributeError("'dict' object has no attribute 'shape'")
    """
    import tensorflow_datasets as tfds

    from flex.data import PluggableDatasetsTensorFlowText

    valid_datasets = []
    wrong_datasets = []
    for dataset in PluggableDatasetsTensorFlowText:
        name, X_columns, y_column = dataset.value
        print(f"Testing dataset: {name}")
        try:
            split = "train" if name != "asset" else "validation"
            data = tfds.load(name, split=split)
            flex_dataset = FedDataDistribution.from_config_with_tfds_dataset_args(
                data, config, X_columns, y_column
            )
            fld = Dataset.from_tfds_dataset_with_args(data, X_columns, y_column)
            fld.validate()
            flex_dataset = FedDataDistribution.from_config(fld, config)
            del flex_dataset
            del data
            del fld
            valid_datasets.append([name, "-"])
        except Exception as e:
            wrong_datasets.append([name, e])
    return valid_datasets, wrong_datasets


def check_if_can_load_torchvision_dataset(list_datasets, **kwargs):
    """Function that recieve a list of datasets and check whether or not
    can be loaded to FLEXible with the methods available. For those that
    gives error, we keep the error.

    Args:
        list_datasets (list): List of list (name, func) containing the datasets
        that will be tested.

    Returns:
        list, list: The first list contains the names of each database that
        can be loaded into FLEXible. The second list contains the names of
        each database that gives error while trying to load it to FLEXible.

    Raises:
        - Gives error if the database can't be loaded to FLEXible.
        - Gives error if the database can't be loaded as: func(split='train')
    """
    valid_datasets = []
    wrong_datasets = []
    for name, func in list_datasets:
        print(f"Testing dataset: {name}")
        try:
            data = func(**kwargs)
            fld = Dataset.from_torchvision_dataset(data)
            fld.validate()
            flex_dataset = FedDataDistribution.from_config(fld, config)
            del flex_dataset
            gc.collect()
            flex_dataset = FedDataDistribution.from_config_with_torchvision_dataset(
                data, config
            )
            del flex_dataset
            del data
            del fld
            valid_datasets.append([name, func, "-"])
        except Exception as e:
            wrong_datasets.append([name, func, e])
        gc.collect()
    return valid_datasets, wrong_datasets



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

"""File that contains the primitive functions to build an easy training loop of the federated learning model.

In this file we specify some functions for each framework, i.e., TensorFlow (tf), PyTorch (pt), among others, but
we only give functions for a general purpose. For a more personalized use of FLEXible, the user must create
her own functions. The user can use this functions as template on how to create a custom function for each step
of the training steps in a federated learning environment.

Note that each function is using the decorators we've created to facilitate the use of the library. For a better
understanding on how the platform works, please go to the flex_decorators file.
"""
from copy import deepcopy  # noqa: E402

from flex.pool.decorators import (  # noqa: E402
    collect_clients_weights,
    deploy_server_model,
    init_server_model,
    set_aggregated_weights,
)


@init_server_model
def init_server_model_tf(
    model=None, optimizer=None, loss=None, metrics=None, *args, **kwargs
):
    """Function that initialize a model in the server side for the TensorFlow framework.

    This function acts as a message handler, that will initialize
    the model at the server side in a client-server architecture.

    Args:
    -----
        model (tf.keras.Model): A tf.keras.model initialized.
        optimizer (tf.keras.optimizers, optional): Optimizer for the model. Defaults to None.
        loss (tf.keras.losses, optional): _description_. Defaults to None.
        metrics (tf.keras.metrics, optional): _description_. Defaults to None.

    Raises:
    -------
        ValueError: If the model is not compiled and any of the optimizer, loss or metrics
        is not provided, then it will raise an error because we can't initialize
        the model.

    Returns:
    --------
        FlexModel: A FlexModel that will be assigned to the server.
    """
    from flex.model.model import FlexModel

    server_flex_model = FlexModel()

    if model._is_compiled:
        server_flex_model["optimizer"] = deepcopy(model.optimizer)
        server_flex_model["loss"] = deepcopy(model.loss)
        server_flex_model["metrics"] = deepcopy(model.compiled_metrics._metrics)
        server_flex_model["model"] = model
    else:
        if any([optimizer, loss, metrics] is None):
            raise ValueError(
                "If the model is not compiled, then optimizer, loss and metrics can't be None. Please, provide a "
                "value for these arguments. "
            )
        server_flex_model["optimizer"] = optimizer
        server_flex_model["loss"] = loss
        server_flex_model["metrics"] = metrics
        server_flex_model["model"] = model
        # Compile the model
        server_flex_model["model"].compile(
            optimizer=server_flex_model["optimizer"],
            loss=server_flex_model["loss"],
            metrics=server_flex_model["metrics"],
        )
    return server_flex_model


@deploy_server_model
def deploy_server_model_tf(server_flex_model, *args, **kwargs):
    """Function to deploy a TensorFlow model from the server to a client.

    The function will make a deepcopy for a TensorFlow model, as it needs
    a special method of copying. Also, it compiles the model for being able
    to train the model.

    This function uses the decorator @deploy_server_model to deploy the
    server_flex_model to the all the clients, so we only need to create
    the steps for 1 client.

    Args:
    -----
        server_flex_model (FlexModel): Server FlexModel

    Returns:
    --------
        FlexModel: The client's FlexModel
    """
    import tensorflow as tf

    from flex.model.model import FlexModel

    weights = server_flex_model["model"].get_weights()
    model = tf.keras.models.clone_model(server_flex_model["model"])
    model.set_weights(weights)
    model.compile(
        optimizer=server_flex_model["optimizer"],
        loss=server_flex_model["loss"],
        metrics=server_flex_model["metrics"],
    )
    client_flex_model = FlexModel()
    client_flex_model["model"] = model
    return client_flex_model


def train_tf(client_flex_model, client_data, *args, **kwargs):
    """Function of general purpose to train a TensorFlow model
    using FLEXible.

    Args:
    -----
        client_flex_model (FlexModel): client's FlexModel
        client_data (FedDataset): client's FedDataset

    Example of use assuming you are using a client-server architecture:

        from flex.pool import train_tf

        clients = flex_pool.clients

        clients.map(train_tf)

    Example of using the FlexPool without separating clients
    and following a client-server architecture.

        from flex.pool import train_tf

        flex_pool.clients.map(train_tf)
    """
    X, y = client_data.to_numpy()
    client_flex_model["model"].fit(X, y, *args, **kwargs)


@collect_clients_weights
def collect_clients_weights_tf(client_flex_model, *args, **kwargs):
    """Function that collect the weights for a TensorFlow model.

    This function returns all the weights of the model.

    Args:
    -----
        client_flex_model (FlexModel): A client's FlexModel

    Returns:
    --------
        np.array: An array with all the weights of the client's model

    Example of use assuming you are using a client-server architecture:

        from flex.pool import collect_weights_tf

        clients = flex_pool.clients
        aggregator = flex_pool.aggregators

        clients.map(collect_weights_tf, aggregator)

    Example of using the FlexPool without separating clients
    and aggregator, and following a client-server architecture.

        from flex.pool import collect_weights_tf

        flex_pool.clients.map(collect_weights_tf, flex_pool.aggregators)
    """
    return client_flex_model["model"].get_weights()


@set_aggregated_weights
def set_aggregated_weights_tf(server_flex_model, aggregated_weights, *args, **kwargs):
    """Function that replaces the weights of the server with the aggregated weights of the aggregator.

    Args:
    -----
        server_flex_model (FlexModel): The server's FlexModel
        aggregated_weights (np.array): An array with the aggregated
        weights of the models.
    """
    server_flex_model["model"].set_weights(aggregated_weights)


def evaluate_model_tf(flex_model, test_data):
    """Function that evaluate the global model on the test data.

    Args:
    -----
        flex_model (FlexModel): server's FlexModel
        test_data (Dataset): Test inputs.

    Returns:
    --------
        Evaluations by the model on the test data.
    """
    X, y = test_data.to_numpy()
    return flex_model["model"].evaluate(X, y, verbose=False)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tensor.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0ctensor.proto\x12\x08\x66lexible\"4\n\x06Tensor\x12\r\n\x05shape\x18\x01 \x03(\x05\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\x12\r\n\x05\x64type\x18\x03 \x01(\tb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tensor_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_TENSOR']._serialized_start=26
  _globals['_TENSOR']._serialized_end=78
# @@protoc_insertion_point(module_scope)



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import collections
from typing import Hashable


class FlexModel(collections.UserDict):
    """Class that represents a model owned by each node in a Federated Experiment.
    The model must be storaged using the key 'model' because of the decorators,
    as we assume that the model will be using that key.
    """

    __actor_id: Hashable = None

    @property
    def actor_id(self):
        return self.__actor_id

    @actor_id.setter
    def actor_id(self, value):
        if self.__actor_id is None:
            self.__actor_id = value
        else:
            raise PermissionError("The property actor_id cannot be updated.")

    @actor_id.deleter
    def actor_id(self):
        ...



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from flex.pool.pool import FlexPool

from flex.pool.primitives_tf import init_server_model_tf
from flex.pool.primitives_tf import deploy_server_model_tf
from flex.pool.primitives_tf import collect_clients_weights_tf
from flex.pool.primitives_tf import train_tf
from flex.pool.primitives_tf import set_aggregated_weights_tf
from flex.pool.primitives_tf import evaluate_model_tf

from flex.pool.primitives_pt import deploy_server_model_pt
from flex.pool.primitives_pt import collect_clients_weights_pt
from flex.pool.primitives_pt import set_aggregated_weights_pt
from flex.pool.primitives_pt import set_aggregated_diff_weights_pt
from flex.pool.primitives_pt import collect_client_diff_weights_pt

from flex.pool.aggregators import fed_avg
from flex.pool.aggregators import fed_avg_f
from flex.pool.aggregators import weighted_fed_avg
from flex.pool.aggregators import weighted_fed_avg_f
from flex.pool.aggregators import set_tensorly_backend

from flex.pool.decorators import init_server_model
from flex.pool.decorators import deploy_server_model
from flex.pool.decorators import collect_clients_weights
from flex.pool.decorators import aggregate_weights
from flex.pool.decorators import set_aggregated_weights
from flex.pool.decorators import evaluate_server_model



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import unittest

import numpy as np
import pytest

from flex.data.dataset import Dataset


@pytest.fixture(name="fcd")
def fixture_simple_fex_data_object():
    X_data = np.random.rand(100).reshape([20, 5])
    y_data = np.random.choice(2, 20)
    return Dataset.from_array(X_data, y_data)


def create_simple_torch_dataset():
    from torch.utils.data import Dataset as TorchDataset

    class SimpleTorchDataset(TorchDataset):
        def __init__(self):
            self.a = np.arange(10)

        def __len__(self):
            return len(self.a)

        def __getitem__(self, idx):
            return self.a[idx], self.a[idx] + 1

    return SimpleTorchDataset()


class TestFlexDataObject(unittest.TestCase):
    @pytest.fixture(autouse=True)
    def _fixture_simple_flex_data_object(self, fcd):
        self._fcd = fcd

    def test_X_data_property(self):
        X_data = np.random.rand(100).reshape([20, 5])
        with pytest.raises(AttributeError):
            self._fcd.X_data = X_data

    def test_y_data_property(self):
        y_data = np.random.choice(2, 20)
        with pytest.raises(AttributeError):
            self._fcd.y_data = y_data

    def test_len_property(self):
        X_data = np.random.rand(100).reshape([20, 5])
        assert len(self._fcd) == len(X_data)

    def test_getitem_property(self):
        X_data = np.random.rand(100).reshape([20, 5])
        y_data = np.random.choice(2, 20)
        fcd = Dataset.from_array(X_data, y_data)
        for x, y, (x_bis, y_bis) in zip(X_data, y_data, fcd):
            assert np.array_equal(x, x_bis)
            assert y == y_bis

    def test_getitem_property_y_data_none(self):
        X_data = np.random.rand(100).reshape([20, 5])
        fcd = Dataset.from_array(X_data)
        for x, (x_bis, y_bis) in zip(X_data, fcd):
            assert np.array_equal(x, x_bis)
            assert y_bis is None

    def test_validate_correct_object(self):
        self._fcd.validate()

    def test_len_X_data_differs_len_y_data(self):
        X_data = np.random.rand(100).reshape([20, 5])
        y_data = np.random.choice(2, 19)
        fcd = Dataset.from_array(X_data, y_data)
        with pytest.raises(ValueError):
            fcd.validate()

    def test_validate_from_torchtext_dataset(self):
        from torchtext.datasets import AG_NEWS

        data = AG_NEWS(split="train")
        fcd = Dataset.from_torchtext_dataset(data)
        fcd.validate()

    def test_validate_from_huggingface_dataset(self):
        from datasets import load_dataset

        data = load_dataset("ag_news", split="train")
        X_columns = ["text"]
        label_columns = ["label"]
        fcd = Dataset.from_huggingface_dataset(
            data, X_columns=X_columns, label_columns=label_columns
        )
        fcd.validate()

    def test_validate_from_huggingface_dataset_lazy(self):
        from datasets import load_dataset

        data = load_dataset("ag_news", split="train")
        X_columns = ["text"]
        label_columns = ["label"]
        fcd = Dataset.from_huggingface_dataset(
            data, X_columns=X_columns, label_columns=label_columns
        )
        fcd.validate()

    def test_validate_from_huggingface_dataset_lazy_with_str_no_subset(self):
        data = "ag_news;train"
        X_columns = ["text"]
        label_columns = ["label"]
        fcd = Dataset.from_huggingface_dataset(
            data, X_columns=X_columns, label_columns=label_columns
        )
        fcd.validate()

    def test_validate_from_huggingface_dataset_lazy_with_str_and_subset(self):
        data = "tweet_eval;emoji;train"
        X_columns = ["text"]
        label_columns = ["label"]
        fcd = Dataset.from_huggingface_dataset(
            data, X_columns=X_columns, label_columns=label_columns
        )
        fcd.validate()

    def test_to_torchvision_dataset_w_flex_datasets(self):
        import torch
        from torchvision import transforms

        from flex.datasets import load

        fcd, _ = load("emnist", split="digits")
        torch_fcd = fcd.to_torchvision_dataset(
            transform=transforms.ToTensor(),
            target_transform=transforms.Compose(
                [
                    lambda x: torch.as_tensor(x).long(),
                    lambda x: torch.nn.functional.one_hot(x, 10),
                ]
            ),
        )
        batch_size = 64
        dataloader = torch.utils.data.DataLoader(
            torch_fcd, batch_size=batch_size, shuffle=True
        )
        train_features, train_labels = next(iter(dataloader))

        assert len(train_features) == batch_size
        assert torch.is_tensor(train_features)
        assert len(train_labels) == batch_size
        assert torch.is_tensor(train_labels)
        assert len(train_labels[0]) == 10  # number of labels

    def test_to_tf_dataset_w_flex_datasets(self):
        import tensorflow as tf

        from flex.datasets import load

        fcd, _ = load("emnist", split="digits")
        tf_fcd = fcd.to_tf_dataset()
        batch_size = 64
        tf_fcd = tf_fcd.batch(batch_size)
        train_features, train_labels = next(iter(tf_fcd))

        assert len(train_features) == batch_size
        assert tf.is_tensor(train_features)
        assert len(train_labels) == batch_size
        assert tf.is_tensor(train_labels)

    def test_pluggable_datasets_in_property(self):
        from torchtext.datasets import AG_NEWS
        from torchvision.datasets import MNIST

        from flex.data.pluggable_datasets import (
            PluggableTorchtext,
            PluggableTorchvision,
        )

        assert AG_NEWS.__name__ in PluggableTorchtext
        assert "random_string" not in PluggableTorchtext
        assert MNIST.__name__ in PluggableTorchvision
        assert "random_string" not in PluggableTorchvision

    def test_from_array_arrays(self):
        X_data = np.random.rand(100).reshape([20, 5])
        y_data = np.random.choice(2, 20)
        fcd = Dataset.from_array(X_data, y_data)
        assert len(fcd.X_data) == len(self._fcd.X_data)
        assert len(fcd.y_data) == len(self._fcd.y_data)

    def test_from_array_arrays_y_none(self):
        X_data = np.random.rand(100).reshape([20, 5])
        fcd = Dataset.from_array(X_data)
        assert len(fcd.X_data) == len(self._fcd.X_data)

    def test_from_arrays(self):
        X_data = list(np.random.rand(100).reshape([20, 5]))
        y_data = list(np.random.choice(2, 20))
        fcd = Dataset.from_array(X_data, y_data)
        assert len(fcd.X_data) == len(self._fcd.X_data)
        assert len(fcd.y_data) == len(self._fcd.y_data)

    def test_from_array_y_none(self):
        X_data = list(np.random.rand(100).reshape([20, 5]))
        fcd = Dataset.from_array(X_data)
        assert len(fcd.X_data) == len(self._fcd.X_data)

    def test_to_list(self):
        X_data, y_data = self._fcd.to_list()
        assert isinstance(X_data, list)
        assert isinstance(y_data, list)
        assert len(X_data) == len(self._fcd.X_data)
        assert len(y_data) == len(self._fcd.y_data)

    def test_to_list_y_none(self):
        X_data = np.random.rand(100).reshape([20, 5])
        fcd = Dataset.from_array(X_data)
        X_data = fcd.to_list()
        assert isinstance(X_data, list)
        assert len(X_data) == len(self._fcd.X_data)

    def test_to_numpy(self):
        X_data, y_data = self._fcd.to_numpy(x_dtype=np.int16, y_dtype=np.int16)
        assert isinstance(X_data, np.ndarray)
        assert isinstance(y_data, np.ndarray)
        assert len(X_data) == len(self._fcd.X_data)
        assert len(y_data) == len(self._fcd.y_data)

    def test_to_numpy_no_dtype(self):
        X_data, y_data = self._fcd.to_numpy()
        assert isinstance(X_data, np.ndarray)
        assert isinstance(y_data, np.ndarray)
        assert len(X_data) == len(self._fcd.X_data)
        assert len(y_data) == len(self._fcd.y_data)

    def test_to_numpy_y_none(self):
        X_data = np.random.rand(100).reshape([20, 5])
        fcd = Dataset.from_array(X_data)
        X_data = fcd.to_numpy(x_dtype=np.int16)
        assert isinstance(X_data, np.ndarray)
        assert len(X_data) == len(self._fcd.X_data)

    def test_from_torchvision_back_to_torchvision(self):
        original_torch_dataset = create_simple_torch_dataset()
        fcd = Dataset.from_torchvision_dataset(original_torch_dataset)
        new_torch_dataset = fcd.to_torchvision_dataset()
        for (x, y), (x_bis, y_bis) in zip(original_torch_dataset, new_torch_dataset):
            assert np.array_equal(x, x_bis)
            assert np.array_equal(y, y_bis)

    def test_from_torchvision_back_to_torchvision_does_not_use_storage(self):
        original_torch_dataset = create_simple_torch_dataset()
        fcd = Dataset.from_torchvision_dataset(original_torch_dataset)
        new_torch_dataset = fcd.to_torchvision_dataset()
        # Consume the datasets
        for (x, y), (x_bis, y_bis) in zip(original_torch_dataset, new_torch_dataset):
            assert np.array_equal(x, x_bis)
            assert np.array_equal(y, y_bis)
        assert fcd.X_data._storage == {}
        assert fcd.y_data._storage == {}



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class Tensor(_message.Message):
    __slots__ = ("shape", "data", "dtype")
    SHAPE_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    DTYPE_FIELD_NUMBER: _ClassVar[int]
    shape: _containers.RepeatedScalarFieldContainer[int]
    data: bytes
    dtype: str
    def __init__(self, shape: _Optional[_Iterable[int]] = ..., data: _Optional[bytes] = ..., dtype: _Optional[str] = ...) -> None: ...



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import unittest

import pytest

from flex.data import Dataset, FedDataDistribution, FedDatasetConfig
from flex.pool.aggregators import fed_avg
from flex.pool.pool import FlexPool
from flex.pool.primitives_tf import (
    collect_clients_weights_tf,
    deploy_server_model_tf,
    evaluate_model_tf,
    init_server_model_tf,
    set_aggregated_weights_tf,
    train_tf,
)


class TestFlexPoolPrimitives(unittest.TestCase):
    @pytest.fixture(autouse=True)
    def _fixture_ag_news_dataset(self):
        from datasets import load_dataset

        config = FedDatasetConfig(
            seed=0,
            n_nodes=2,
            replacement=False,
            node_ids=["client_0", "client_1"],
        )
        self.config = config
        train_data, test_data = load_dataset("imdb", split=["train", "test"])
        self.test_data = test_data
        X_columns = ["text"]
        label_columns = ["label"]
        self.f_imdb = FedDataDistribution.from_config_with_huggingface_dataset(
            train_data, self.config, X_columns, label_columns
        )
        self.f_imdb["server"] = Dataset.from_huggingface_dataset(
            self.test_data, X_columns, label_columns
        )

    def test_primitives_tf(self):
        import numpy as np
        import tensorflow as tf
        import tensorflow_hub as hub

        def define_model(*args):
            model = "https://tfhub.dev/google/nnlm-en-dim50/2"
            hub_layer = hub.KerasLayer(
                model, input_shape=[], dtype=tf.string, trainable=True
            )
            model = tf.keras.Sequential()
            model.add(hub_layer)
            model.add(tf.keras.layers.Dense(16, activation="relu"))
            model.add(tf.keras.layers.Dense(1))
            model.compile(
                optimizer="adam",
                loss=tf.losses.BinaryCrossentropy(from_logits=True),
                metrics=[tf.metrics.BinaryAccuracy(threshold=0.0, name="accuracy")],
            )
            return model

        p = FlexPool.client_server_pool(
            self.f_imdb, init_func=init_server_model_tf, model=define_model()
        )
        reference_model = define_model()
        reference_model_params = reference_model.get_weights()
        p.servers.map(deploy_server_model_tf, p.clients)
        assert all(
            np.all(
                np.equal(
                    p._models[k]["model"].get_weights()[0], reference_model_params[0]
                )
            )
            for k in p.actor_ids
        )
        # Train the model
        p.clients.map(train_tf, batch_size=512, epochs=1)
        # Collect weights
        p.aggregators.map(collect_clients_weights_tf, p.clients)
        # Aggregate weights
        p.aggregators.map(fed_avg)
        assert np.any(
            np.not_equal(
                p._models["server"]["aggregated_weights"][0], reference_model_params[0]
            )
        )
        # Transfer weights from aggregators to servers
        p.aggregators.map(set_aggregated_weights_tf, p.servers)
        # Deploy new model to clients
        p.servers.map(deploy_server_model_tf, p.clients)
        reference_value = p._models["server"]["model"].get_weights()
        assert all(
            np.all(np.equal(p._models[k]["model"].get_weights()[0], reference_value[0]))
            for k in p.actor_ids
        )
        result = p.servers.map(evaluate_model_tf)
        print(result)



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

"""File that contains the primitive functions to build an easy training loop of the federated learning model.

In this file we specify some functions for each framework, i.e., TensorFlow (tf), PyTorch (pt), among others, but
we only give functions for a general purpose. For a more personalized use of FLEXible, the user must create
her own functions. The user can use this functions as template on how to create a custom function for each step
of the training steps in a federated learning environment.

Note that each function is using the decorators we've created to facilitate the use of the library. For a better
understanding on how the platform works, please go to the flex_decorators file.
"""
from copy import deepcopy  # noqa: E402

from flex.pool.decorators import (  # noqa: E402
    collect_clients_weights,
    deploy_server_model,
    set_aggregated_weights,
)


@deploy_server_model
def deploy_server_model_pt(server_flex_model, *args, **kwargs):
    """Creates a copy of the server_flex_model and it is set to client nodes using the decorator @deploy_server_model.

    Args:
    -----
        server_flex_model (FlexModel): object storing information needed to run a Pytorch model

    """
    return deepcopy(server_flex_model)


def check_ignored_weights_pt(name, ignore_weights=None):
    """Checks wether name contains any of the words in ignore_weights.

    Args:
    -----
        name (str): name to check
        ignore_weights (list, optional): A list of str. Defaults to None.

    Returns:
    --------
        bool: True if any og the elements of list ignore_weights is present in name, otherwise False.
    """
    if ignore_weights is None:
        ignore_weights = ["num_batches_tracked"]
    return any(ignored in name for ignored in ignore_weights)


@collect_clients_weights
def collect_client_diff_weights_pt(client_flex_model, *args, **kwargs):
    # sourcery skip: raise-specific-error
    """Function that collect the weights for a PyTorch model. Particularly,
        it collects the difference between the model before and after training, \
        that is, what the model has learnt in its local training step. Also note \
        that the weights of the model before training are assume to be stored \
        using `previous_model` as key.

    This function returns the weights of the model.

    Args:
    -----
        client_flex_model (FlexModel): A client's FlexModel
        ignore_weights (list): the name of the weights not to collect, by default,
        those containind the words `num_batches_tracked` are not collected, as they
        only make sense in the local model

    Returns:
    --------
        List: List with the weights of the client's model

    Example of use assuming you are using a client-server architecture:

        from flex.pool import collect_client_diff_weights_pt

        clients = flex_pool.clients
        aggregator = flex_pool.aggregators

        clients.map(collect_client_diff_weights_pt, aggregator)

    Example of using the FlexPool without separating clients
    and aggregator, and following a client-server architecture.

        from flex.pool import collect_client_diff_weights_pt

        flex_pool.clients.map(collect_client_diff_weights_pt, flex_pool.aggregators)
    """
    import torch

    ignore_weights = kwargs.get("ignore_weights", None)
    with torch.no_grad():
        weight_dict = client_flex_model["model"].state_dict()
        try:
            previous_weight_dict = client_flex_model["previous_model"].state_dict()
        except KeyError as e:
            raise Exception(
                'A copy of the model before training must be stored in client FlexModel using key: "previous_model"'
            ) from e
        parameters = []
        for name in weight_dict:
            if check_ignored_weights_pt(name, ignore_weights=ignore_weights):
                parameters.append(torch.tensor([]))
                continue
            weight_diff = weight_dict[name] - previous_weight_dict[name]
            parameters.append(weight_diff)
    return parameters


@collect_clients_weights
def collect_clients_weights_pt(client_flex_model, *args, **kwargs):
    """Function that collect the weights for a PyTorch model.

    This function returns all the weights of the model.

    Args:
    -----
        client_flex_model (FlexModel): A client's FlexModel
        ignore_weights (list): the name of the weights not to collect, by default,
        those containind the words `num_batches_tracked` are not collected, as they
        only make sense in the local model

    Returns:
    --------
        List: List with all the weights of the client's model

    Example of use assuming you are using a client-server architecture:

        from flex.pool import collect_weights_pt

        clients = flex_pool.clients
        aggregator = flex_pool.aggregators

        clients.map(collect_weights_pt, aggregator)

    Example of using the FlexPool without separating clients
    and aggregator, and following a client-server architecture.

        from flex.pool import collect_weights_pt

        flex_pool.clients.map(collect_weights_pt, flex_pool.aggregators)
    """
    import torch

    ignore_weights = kwargs.get("ignore_weights", None)
    with torch.no_grad():
        parameters = []
        weight_dict = client_flex_model["model"].state_dict()
        for name in weight_dict:
            w = weight_dict[name]
            if check_ignored_weights_pt(name, ignore_weights=ignore_weights):
                w = torch.tensor([])
                continue
            parameters.append(w)
    return parameters


@set_aggregated_weights
def set_aggregated_weights_pt(server_flex_model, aggregated_weights, *args, **kwargs):
    """Function that replaces the weights of the server with the aggregated weights of the aggregator.

    Args:
    -----
        server_flex_model (FlexModel): The server's FlexModel
        aggregated_weights (np.array): Aggregated weights

    Example of use assuming you are using a client-server architecture:

        from flex.pool import set_aggregated_weights_pt

        aggregator = flex_pool.aggregators

        aggregator.map(set_aggregated_weights_pt)

    Example of using the FlexPool without separating clients
    and aggregator, and following a client-server architecture.

        from flex.pool import set_aggregated_weights_pt

        flex_pool.aggregators.map(set_aggregated_weights_pt)
    """
    import torch

    with torch.no_grad():
        weight_dict = server_flex_model["model"].state_dict()
        for layer_key, new in zip(weight_dict, aggregated_weights):
            try:
                if len(new) != 0:  # Do not copy empty layers
                    weight_dict[layer_key].copy_(new)
            except TypeError:  # new has no len property
                weight_dict[layer_key].copy_(new)


@set_aggregated_weights
def set_aggregated_diff_weights_pt(
    server_flex_model, aggregated_diff_weights, *args, **kwargs
):
    """Function to add the aggregated weights to the server.

    Args:
    -----
        server_flex_model (FlexModel): The server's FlexModel
        aggregated_diff_weights (np.array): Aggregated weights

    Example of use assuming you are using a client-server architecture:

        from flex.pool import set_aggregated_diff_weights_pt

        aggregator = flex_pool.aggregators

        aggregator.map(set_aggregated_diff_weights_pt)

    Example of using the FlexPool without separating clients
    and aggregator, and following a client-server architecture.

        from flex.pool import set_aggregated_diff_weights_pt

        flex_pool.aggregators.map(set_aggregated_diff_weights_pt)
    """
    import torch

    with torch.no_grad():
        weight_dict = server_flex_model["model"].state_dict()
        for layer_key, new in zip(weight_dict, aggregated_diff_weights):
            try:
                if len(new) != 0:  # Do not copy empty layers
                    weight_dict[layer_key].add_(new)
            except TypeError:  # new has no len property
                weight_dict[layer_key].add_(new)



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from inspect import getmembers, isfunction

from flex.datasets import federated_datasets, standard_datasets


def list_datasets():
    """
    Function that returns the available datasets in FLEXible.
    Returns: A list with the available datasets.
    -------

    """
    return list(ds_invocation.keys())


def load(name, *args, **kwargs):
    """
    Function that loads every dataset builtin into the framework
    Parameters
    ----------
    name


    Returns
    -------

    """
    try:
        caller = ds_invocation[name.lower()]
    except KeyError as e:
        raise ValueError(
            f"{name} is not available. Please, check available datasets using list_datasets "
            f"function. "
        ) from e

    return caller(*args, **kwargs)


ds_invocation = dict(
    getmembers(standard_datasets, isfunction)
    + getmembers(federated_datasets, isfunction)
)



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from collections import OrderedDict
from typing import Hashable

from flex.actors import FlexRole, FlexRoleManager


class FlexActors(OrderedDict):
    """Class that represents roles assigned to each node in a Federated Experiment.
    Roles are designed to restrict communications between nodes. It is important
    to note that Roles are not mutually exclusive, that is, a node can have multiple
    Roles.

    Attributes
    ----------
    data (collections.OrderedDict): The structure is a dictionary
        with the node ids as keys and Roles as a value.
    """

    def check_compatibility(self, key1: Hashable, key2: Hashable) -> bool:
        """Method to ensure that it is possible to establish communication
        between two actors, according to their roles. Note that the communication
        is stablished from node with key1 to node with key2. Communication from node
        with key2 to node with key1 is not checked.

        Args:
        -----
            key1 (Hashable): id used to identify a node. This node is supposed to start communication
            from itself to node with key2.
            key2 (Hashable): id used to identify a node. This node is suppored to receive communication
            from node with key1.

        Returns:
        --------
            bool: whether or not the communication is allowed.
        """
        return FlexRoleManager.check_compatibility(self[key1], self[key2])

    def __setitem__(self, key: Hashable, item: FlexRole) -> None:
        super().__setitem__(key, item)



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import unittest
from unittest import mock

import numpy as np
import pytest

from flex.data import Dataset
from flex.distributed import ClientBuilder
from flex.distributed.common import (  # required for internal testing
    toNumpyArray,
    toTensorList,
)
from flex.distributed.proto.transport_pb2 import (  # required for internal testing
    ServerMessage,
)
from flex.model import FlexModel
from flex.pool.decorators import (
    collect_clients_weights,
    init_server_model,
    set_aggregated_weights,
)


@init_server_model
def build_server_model():
    model = FlexModel()
    model["model"] = {"weights1": [1, 2, 3], "weights2": [4, 5, 6]}
    return model


@collect_clients_weights
def collect_weights(model):
    return [model["model"]["weights1"], model["model"]["weights2"]]


@set_aggregated_weights
def set_weights(model, weights):
    model["model"]["weights1"] = weights[0]
    model["model"]["weights2"] = weights[1]


def train(model, data):
    model["model"]["weights1"] = [7, 8, 9]


def eval(model, data):
    return {"accuracy": 0.5}


class TestClientBuilder(unittest.TestCase):
    @pytest.fixture(autouse=True)
    def _fixture_dataset(self):
        self.dataset = Dataset.from_array([[1, 2], [3, 4]], [0, 1])

    def test_builder_raises_when_not_all_fields_are_called(self):
        with pytest.raises(AssertionError):
            ClientBuilder().collect_weights(collect_weights).set_weights(
                set_weights
            ).train(train).build()

    def test_builder_builds_working_client(self):
        client = (
            ClientBuilder()
            .collect_weights(collect_weights)
            .build_model(build_server_model)
            .set_weights(set_weights)
            .train(train)
            .eval(eval, self.dataset)
            .dataset(self.dataset)
            .build()
        )

        mock_stub = mock.Mock()
        mock_stub.Send = mock.Mock(return_value=mock.MagicMock())
        iter_mock = mock_stub.Send.return_value
        iter_mock.__iter__ = mock.Mock(
            return_value=iter(
                [
                    ServerMessage(
                        get_weights_ins=ServerMessage.GetWeightsIns(status=200)
                    ),
                    ServerMessage(train_ins=ServerMessage.TrainIns(status=200)),
                    ServerMessage(eval_ins=ServerMessage.EvalIns(status=200)),
                    ServerMessage(
                        get_weights_ins=ServerMessage.GetWeightsIns(status=200)
                    ),
                    ServerMessage(
                        send_weights_ins=ServerMessage.SendWeightsIns(
                            weights=toTensorList(np.array([[1, 2], [3, 4]]))
                        )
                    ),
                    ServerMessage(
                        get_weights_ins=ServerMessage.GetWeightsIns(status=200)
                    ),
                ]
            )
        )

        client.run("localhost:50051", _stub=mock_stub)

        # Obtain the generator passed to the stub where the client sends messages
        mock_stub.Send.assert_called_once()
        generator = mock_stub.Send.mock_calls[0].args[0]

        # Always start with handshake
        response = next(generator)
        self.assertEqual(response.handshake_res.status, 200)
        # obtain initial weights
        response = next(generator)
        weights = toNumpyArray(response.get_weights_res.weights)
        self.assertEqual([w.tolist() for w in weights], [[1, 2, 3], [4, 5, 6]])
        # train
        response = next(generator)
        # eval
        response = next(generator)
        self.assertEqual(response.eval_res.metrics["accuracy"], 0.5)
        # see if weights changed
        response = next(generator)
        weights = toNumpyArray(response.get_weights_res.weights)
        self.assertEqual([w.tolist() for w in weights], [[7, 8, 9], [4, 5, 6]])
        # send weights
        response = next(generator)
        self.assertEqual(response.send_weights_res.status, 200)
        # get weights again
        response = next(generator)
        weights = toNumpyArray(response.get_weights_res.weights)
        self.assertEqual([w.tolist() for w in weights], [[1, 2], [3, 4]])

        # The StopIteration in the generator raises a RuntimeError
        self.assertRaises(RuntimeError, lambda: next(generator))



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import os
import unittest
from math import isclose

import numpy as np
import pytest
from sklearn.cluster import KMeans
from sklearn.datasets import load_iris

from flex.data import Dataset, FedDataDistribution, FedDataset, FedDatasetConfig
from flex.datasets import load


@pytest.fixture(name="fcd")
def fixture_simple_fex_data_object():
    X_data = np.random.rand(100).reshape([20, 5])
    y_data = np.random.choice(2, 20)
    return Dataset.from_array(X_data, y_data)


@pytest.fixture(name="fcd_ones_zeros")
def fixture_simple_fex_data_object_ones_zeros():
    X_data = np.random.rand(100).reshape([20, 5])
    y_data = np.concatenate((np.zeros(10), np.ones(10)))
    return Dataset.from_array(X_data, y_data)


@pytest.fixture(name="fcd_multiple_classes")
def fixture_simple_fex_data_object_multiple_classes():
    X_data = np.random.rand(150).reshape([30, 5])
    y_data = np.concatenate(
        (
            np.zeros(5),
            np.ones(5),
            2 * np.ones(5),
            3 * np.ones(5),
            4 * np.ones(5),
            5 * np.ones(5),
        )
    )
    return Dataset.from_array(X_data, y_data)


class TestFlexDataDistribution(unittest.TestCase):
    @pytest.fixture(autouse=True)
    def _fixture_iris_dataset(self):
        iris = load_iris()
        self._iris = Dataset.from_array(iris.data, iris.target)

    @pytest.fixture(autouse=True)
    def _fixture_simple_flex_data_object(self, fcd):
        self._fcd = fcd

    @pytest.fixture(autouse=True)
    def _fixture_simple_flex_data_object_ones_zeros(self, fcd_ones_zeros):
        self._fcd_ones_zeros = fcd_ones_zeros

    @pytest.fixture(autouse=True)
    def _fixture_simple_flex_data_object_multiple_classes(self, fcd_multiple_classes):
        self._fcd_multiple_classes = fcd_multiple_classes

    def test_init_method_does_not_work(self):
        with pytest.raises(AssertionError):
            FedDataDistribution()

    def test_client_names(self):
        config = FedDatasetConfig(n_nodes=3, node_ids=["Juan", "Pepe", 2])
        flex_dataset = FedDataDistribution.from_config(self._fcd, config)
        assert "Juan" in flex_dataset
        assert "Pepe" in flex_dataset
        assert 2 in flex_dataset

    def test_client_names_bis(self):
        config = FedDatasetConfig(n_nodes=3, node_ids=["Juan", "Pepe", 2])
        flex_dataset = FedDataDistribution.from_config(self._fcd, config)
        assert "Juan" in flex_dataset
        assert "Pepe" in flex_dataset
        assert 2 in flex_dataset

    def test_nclients(self):
        config = FedDatasetConfig(n_nodes=2)
        flex_dataset = FedDataDistribution.from_config(self._fcd, config)
        assert len(flex_dataset) == config.n_nodes
        assert len(flex_dataset[0]) == len(flex_dataset[1])
        assert len(flex_dataset[0]) + len(flex_dataset[1]) == len(self._fcd)
        assert (
            set(flex_dataset[0].X_data._iterable_indexes)
            & set(flex_dataset[1].X_data._iterable_indexes)
            == set()
        )

    def test_shuffle_true(self):
        config = FedDatasetConfig(
            seed=0, n_nodes=2, weights=[1, 1], replacement=True, shuffle=True
        )
        flex_dataset = FedDataDistribution.from_config(self._iris, config)
        assert len(flex_dataset) == config.n_nodes
        assert len(flex_dataset[0]) == len(flex_dataset[1])
        assert len(flex_dataset[0]) == len(self._iris)
        assert set(flex_dataset[0].X_data._iterable_indexes) == set(
            flex_dataset[1].X_data._iterable_indexes
        )
        assert any(flex_dataset[0].X_data[0] != flex_dataset[1].X_data[0])

    def test_shuffle_false(self):
        config = FedDatasetConfig(
            seed=0, n_nodes=2, weights=[1, 1], replacement=True, shuffle=False
        )
        flex_dataset = FedDataDistribution.from_config(self._iris, config)
        assert len(flex_dataset) == config.n_nodes
        assert len(flex_dataset[0]) == len(flex_dataset[1])
        assert len(flex_dataset[0]) + len(flex_dataset[1]) == 2 * len(self._iris)
        assert set(flex_dataset[0].X_data._iterable_indexes) == set(
            flex_dataset[1].X_data._iterable_indexes
        )
        assert all(flex_dataset[0].X_data[0] == flex_dataset[1].X_data[0])

    def test_weights(self):
        config = FedDatasetConfig(n_nodes=2, weights=[1, 1], replacement=False)
        flex_dataset = FedDataDistribution.from_config(self._fcd, config)
        assert len(flex_dataset) == config.n_nodes
        assert len(flex_dataset[0]) == len(flex_dataset[1])
        assert len(set(flex_dataset[0].y_data)) == 2
        assert len(set(flex_dataset[1].y_data)) == 2
        assert (
            set(flex_dataset[0].X_data._iterable_indexes)
            & set(flex_dataset[1].X_data._iterable_indexes)
            == set()
        )

    def test_empty_weights(self):
        config = FedDatasetConfig(n_nodes=2, weights=None, replacement=False)
        flex_dataset = FedDataDistribution.from_config(self._fcd, config)
        assert len(flex_dataset) == config.n_nodes
        assert len(flex_dataset[1]) == len(flex_dataset[0])
        assert len(set(flex_dataset[1].y_data)) == 2
        assert len(set(flex_dataset[0].y_data)) == 2
        assert (
            set(flex_dataset[0].X_data._iterable_indexes)
            & set(flex_dataset[1].X_data._iterable_indexes)
            == set()
        )

    def test_classes_per_client_int_no_weights_no_replacement(self):
        config = FedDatasetConfig(
            seed=1, n_nodes=2, labels_per_node=1, replacement=False
        )
        flex_dataset = FedDataDistribution.from_config(self._fcd_ones_zeros, config)
        assert len(flex_dataset) == config.n_nodes
        assert len(flex_dataset[0]) + len(flex_dataset[1]) == len(self._fcd_ones_zeros)
        assert len(set(flex_dataset[0].y_data)) == 1
        assert len(set(flex_dataset[1].y_data)) == 1

    def test_classes_per_client_int_no_weights_with_replacements(self):
        config = FedDatasetConfig(n_nodes=2, labels_per_node=1, replacement=True)
        flex_dataset = FedDataDistribution.from_config(self._fcd_ones_zeros, config)
        assert len(flex_dataset) == config.n_nodes
        assert len(flex_dataset[0]) + len(flex_dataset[1]) == len(self._fcd_ones_zeros)
        assert len(set(flex_dataset[0].y_data)) == 1
        assert len(set(flex_dataset[1].y_data)) == 1

    def test_classes_per_client_int_with_weigths_no_replacement(self):
        config = FedDatasetConfig(
            n_nodes=2, labels_per_node=1, weights=[0.25, 0.5], replacement=False
        )
        flex_dataset = FedDataDistribution.from_config(self._fcd_ones_zeros, config)
        assert len(flex_dataset) == config.n_nodes
        assert len(flex_dataset[0]) == 2
        assert len(flex_dataset[1]) == 5
        assert len(set(flex_dataset[0].y_data)) == 1
        assert len(set(flex_dataset[1].y_data)) == 1

    def test_classes_per_client_int_with_weigths_with_replacement(self):
        config = FedDatasetConfig(
            seed=1,
            n_nodes=2,
            labels_per_node=1,
            weights=[0.25, 0.5],
            replacement=True,
        )
        flex_dataset = FedDataDistribution.from_config(self._fcd_ones_zeros, config)
        assert len(flex_dataset) == config.n_nodes
        assert len(flex_dataset[0]) + len(flex_dataset[1]) == int(
            sum(np.floor(np.array(config.weights) * 10))
        )
        assert len(set(flex_dataset[0].y_data)) == 1
        assert len(set(flex_dataset[1].y_data)) == 1

    def test_classes_per_client_tuple_with_weights_no_replacement(self):
        config = FedDatasetConfig(
            seed=1,
            n_nodes=2,
            labels_per_node=(1, 2),
            weights=[0.5, 0.5],
            replacement=False,
        )

        flex_dataset = FedDataDistribution.from_config(
            self._fcd_multiple_classes, config
        )
        assert len(flex_dataset) == config.n_nodes
        for k in flex_dataset:
            assert len(set(flex_dataset[k].y_data)) <= 2
            assert len(set(flex_dataset[k].y_data)) >= 1

    def test_classes_per_client_tuple_with_weights_with_replacement(self):
        config = FedDatasetConfig(
            seed=1,
            n_nodes=2,
            labels_per_node=(2, 3),
            weights=[0.5, 1],
            replacement=True,
        )

        flex_dataset = FedDataDistribution.from_config(
            self._fcd_multiple_classes, config
        )
        assert len(flex_dataset) == config.n_nodes
        for k in flex_dataset:
            assert len(set(flex_dataset[k].y_data)) <= 3
            assert len(set(flex_dataset[k].y_data)) >= 2

    def test_classes_per_client_arr_no_weights_no_replacement(self):
        config = FedDatasetConfig(
            seed=2,
            n_nodes=2,
            node_ids=["client_0", "client_1"],
            labels_per_node=[[0], [1]],
            weights=None,
            replacement=False,
        )
        flex_dataset = FedDataDistribution.from_config(self._iris, config)
        assert set(flex_dataset["client_0"].y_data) == {0}
        assert set(flex_dataset["client_1"].y_data) == {1}
        # Empty intersection
        assert (
            set(flex_dataset["client_0"].X_data._iterable_indexes)
            & set(flex_dataset["client_1"].X_data._iterable_indexes)
            == set()
        )

    def test_classes_per_client_arr_no_weights_with_replacement(self):
        config = FedDatasetConfig(
            seed=2,
            n_nodes=2,
            node_ids=["client_0", "client_1"],
            labels_per_node=[[0], [0, 1]],
            weights=None,
            replacement=True,
        )
        flex_dataset = FedDataDistribution.from_config(self._iris, config)
        assert len(flex_dataset) == config.n_nodes
        assert set(flex_dataset["client_0"].y_data) == {0}
        assert set(flex_dataset["client_1"].y_data) == {0, 1}
        # Non-Empty intersection
        assert (
            set(flex_dataset["client_0"].X_data._iterable_indexes)
            & set(flex_dataset["client_1"].X_data._iterable_indexes)
            != set()
        )

    def test_classes_per_client_arr_with_weights_with_replacement(self):
        config = FedDatasetConfig(
            seed=2,
            n_nodes=2,
            node_ids=["client_0", "client_1"],
            labels_per_node=[[0], [1, 0]],
            weights=[0.25, 0.5],
            replacement=True,
        )
        flex_dataset = FedDataDistribution.from_config(self._iris, config)
        assert len(flex_dataset) == config.n_nodes
        assert set(flex_dataset["client_0"].y_data) == {0}
        assert set(flex_dataset["client_1"].y_data) == {0, 1}

    def test_classes_per_client_arr_with_weights_no_replacement(self):
        config = FedDatasetConfig(
            seed=2,
            n_nodes=2,
            node_ids=["client_0", "client_1"],
            labels_per_node=[[0], [1]],
            weights=[0.25, 0.5],
            replacement=False,
        )
        flex_dataset = FedDataDistribution.from_config(self._iris, config)
        assert len(flex_dataset) == config.n_nodes
        assert set(flex_dataset["client_0"].y_data) == {0}
        assert set(flex_dataset["client_1"].y_data) == {1}
        # Empty intersection
        assert (
            set(flex_dataset["client_0"].X_data._iterable_indexes)
            & set(flex_dataset["client_1"].X_data._iterable_indexes)
            == set()
        )

    # Feature split testing
    def test_featutes_per_client_int(self):
        config = FedDatasetConfig(
            seed=2, n_nodes=2, features_per_node=3, replacement=True, weights=[1.0, 1.0]
        )
        flex_dataset = FedDataDistribution.from_config(self._iris, config)
        assert len(flex_dataset) == config.n_nodes
        assert len(flex_dataset[0]) == len(flex_dataset[1])
        assert len(flex_dataset[0]) == len(self._iris)
        for k in flex_dataset:
            assert flex_dataset[k].X_data.to_numpy().shape[1] == 3
        # Non-Empty intersection
        assert (
            set(flex_dataset[0].X_data._iterable_indexes)
            & set(flex_dataset[1].X_data._iterable_indexes)
            != set()
        )

    def test_featutes_per_client_tuple(self):
        min_features = 1
        max_features = 3
        config = FedDatasetConfig(
            seed=2,
            n_nodes=2,
            features_per_node=(min_features, max_features),
            replacement=True,
        )
        flex_dataset = FedDataDistribution.from_config(self._iris, config)
        assert len(flex_dataset) == config.n_nodes
        assert len(flex_dataset[0]) == len(flex_dataset[1])
        assert len(flex_dataset[0]) != len(self._iris)
        for k in flex_dataset:
            assert flex_dataset[k].X_data.to_numpy().shape[1] <= max_features
            assert flex_dataset[k].X_data.to_numpy().shape[1] >= min_features

    def test_featutes_per_client_arr(self):
        config = FedDatasetConfig(
            seed=2,
            n_nodes=2,
            node_ids=["client_0", "client_1"],
            features_per_node=[[1, 3], [0, 2]],
            replacement=True,
        )
        flex_dataset = FedDataDistribution.from_config(self._iris, config)
        assert len(flex_dataset) == config.n_nodes
        assert flex_dataset["client_0"].X_data.to_numpy().shape[1] == 2
        assert flex_dataset["client_1"].X_data.to_numpy().shape[1] == 2
        assert len(flex_dataset["client_0"]) == len(flex_dataset["client_1"])
        assert len(flex_dataset["client_0"]) != len(self._iris)
        # Non-Empty intersection
        assert (
            set(flex_dataset["client_0"].X_data._iterable_indexes)
            & set(flex_dataset["client_1"].X_data._iterable_indexes)
            != set()
        )

    def test_iid_distribution(self):
        n_nodes = 2
        flex_dataset = FedDataDistribution.iid_distribution(self._iris, n_nodes)
        assert len(flex_dataset) == n_nodes
        assert len(flex_dataset[0]) == len(flex_dataset[1])
        assert len(flex_dataset[0]) + len(flex_dataset[1]) == len(self._iris)
        assert (
            set(flex_dataset[0].X_data._iterable_indexes)
            & set(flex_dataset[1].X_data._iterable_indexes)
            == set()
        )

    def test_single_feature_data(self):
        new_X_data = self._iris.X_data.to_numpy()
        new_X_data = new_X_data[:, 0]
        single_feature_dataset = Dataset.from_array(list(new_X_data), self._iris.y_data)
        federated_iris = FedDataDistribution.iid_distribution(
            centralized_data=single_feature_dataset
        )
        assert isinstance(federated_iris[0].X_data[0], float)

    def test_indexes_per_client(self):
        indexes = [[1, 3], [0, 2]]
        config = FedDatasetConfig(
            n_nodes=len(indexes), replacement=False, indexes_per_node=indexes
        )
        federated_iris = FedDataDistribution.from_config(self._iris, config)
        assert all(
            np.array_equal(federated_iris[client].X_data, self._iris.X_data[idx])
            and np.array_equal(federated_iris[client].y_data, self._iris.y_data[idx])
            for idx, client in zip(indexes, federated_iris)
        )

    def test_from_clustering_func(self):
        n_nodes = 10
        kmeans = KMeans(n_clusters=n_nodes, random_state=0).fit(self._iris.X_data)
        federated_iris = FedDataDistribution.from_clustering_func(
            self._iris, clustering_func=lambda x, _: kmeans.predict(x.reshape(1, -1))[0]
        )
        assert len(federated_iris) == n_nodes
        for idx, client in enumerate(kmeans.labels_):
            assert self._iris.X_data[idx] in federated_iris[client].X_data.to_numpy()

    def test_weight_per_classes_random_assigment(self):
        classes = np.unique(self._iris.y_data)
        config = FedDatasetConfig(
            seed=2,
            n_nodes=3,
            weights_per_label=[[1, 2, 3], [0, 0, 1], [0, 1, 0]],
            replacement=False,
        )
        flex_dataset = FedDataDistribution.from_config(self._iris, config)
        clients = list(flex_dataset.keys())
        assert all(
            sum(flex_dataset[clients[0]].y_data == i)
            != sum(flex_dataset[clients[1]].y_data == i)
            for i in classes
        )
        assert len(set(flex_dataset[0].y_data)) == 3
        assert len(set(flex_dataset[1].y_data)) == 1
        assert len(set(flex_dataset[2].y_data)) == 1

    def test_weight_per_class_alone_w_replacement(self):
        classes = np.unique(self._iris.y_data)
        config = FedDatasetConfig(
            seed=2,
            n_nodes=2,
            weights_per_label=[[1, 1, 1], [1, 1, 1]],
            replacement=True,
        )
        flex_dataset = FedDataDistribution.from_config(self._iris, config)
        clients = list(flex_dataset.keys())
        assert all(
            sum(flex_dataset[clients[0]].y_data == i)
            == sum(flex_dataset[clients[1]].y_data == i)
            for i in classes
        )
        assert len(set(flex_dataset[0].y_data)) == 3
        assert len(set(flex_dataset[1].y_data)) == 3
        assert (
            set(flex_dataset[0].X_data._iterable_indexes)
            & set(flex_dataset[1].X_data._iterable_indexes)
            != set()
        )

    def test_weight_per_class_alone_without_replacement(self):
        classes = np.unique(self._iris.y_data)
        config = FedDatasetConfig(
            seed=2,
            n_nodes=2,
            weights_per_label=[[1, 1, 1], [1, 1, 1]],
            replacement=False,
        )
        flex_dataset = FedDataDistribution.from_config(self._iris, config)
        clients = list(flex_dataset.keys())
        assert all(
            sum(flex_dataset[clients[0]].y_data == i)
            == sum(flex_dataset[clients[1]].y_data == i)
            for i in classes
        )
        assert len(set(flex_dataset[0].y_data)) == 3
        assert len(set(flex_dataset[1].y_data)) == 3
        assert (
            set(flex_dataset[0].X_data._iterable_indexes)
            & set(flex_dataset[1].X_data._iterable_indexes)
            == set()
        )

    def test_keep_labels(self):
        config = FedDatasetConfig(
            n_nodes=2,
            node_ids=["client_0", "client_1"],
            keep_labels=[True, False],
        )
        flex_dataset = FedDataDistribution.from_config(self._iris, config)
        assert flex_dataset["client_0"].y_data is not None
        assert flex_dataset["client_1"].y_data is None

    def test_from_torchtext_dataset(self):
        from torchtext.datasets import AG_NEWS

        data = AG_NEWS(split="train")
        config = FedDatasetConfig(
            seed=0,
            n_nodes=2,
            replacement=True,
            node_ids=["client_0", "client_1"],
        )
        flex_dataset = FedDataDistribution.from_config_with_torchtext_dataset(
            data, config
        )
        assert len(flex_dataset) == config.n_nodes
        assert len(flex_dataset["client_0"]) == len(flex_dataset["client_1"])

    def test_from_tfds_image_dataset(self):
        import tensorflow_datasets as tfds

        # With batch_size -1
        def build_data_and_check(data, config):
            flex_dataset = FedDataDistribution.from_config_with_tfds_image_dataset(
                data, config
            )
            assert len(flex_dataset) == config.n_nodes
            assert len(flex_dataset["client_0"]) == len(flex_dataset["client_1"])

        other_options = {
            "split": "test",
            "shuffle_files": True,
            "as_supervised": True,
            "batch_size": -1,
        }
        ds_name = "cifar10"
        data = tfds.load(ds_name, **other_options)
        config = FedDatasetConfig(
            seed=0,
            n_nodes=2,
            replacement=True,
            node_ids=["client_0", "client_1"],
        )
        # With batch_size = 20
        other_options["batch_size"] = 20
        data = tfds.load(ds_name, **other_options)
        build_data_and_check(data, config)

    def test_from_tfds_image_dataset_without_batchsize(self):
        import tensorflow_datasets as tfds

        def build_data_and_check(data, config):
            flex_dataset = FedDataDistribution.from_config_with_tfds_image_dataset(
                data, config
            )
            assert len(flex_dataset) == config.n_nodes
            assert len(flex_dataset["client_0"]) == len(flex_dataset["client_1"])

        other_options = {
            "split": "test",
            "shuffle_files": True,
            "as_supervised": True,
            "batch_size": -1,
        }
        ds_name = "cifar10"
        data = tfds.load(ds_name, **other_options)
        config = FedDatasetConfig(
            seed=0,
            n_nodes=2,
            replacement=False,
            node_ids=["client_0", "client_1"],
        )
        data = tfds.load(ds_name, **other_options)
        build_data_and_check(data, config)

    def test_from_tfds_text_dataset(self):
        import tensorflow_datasets as tfds

        def build_data_and_check(data, config, X_columns, labels):
            flex_dataset = FedDataDistribution.from_config_with_tfds_text_dataset(
                data, config, X_columns, labels
            )
            assert len(flex_dataset) == config.n_nodes
            assert len(flex_dataset["client_0"]) == len(flex_dataset["client_1"])

        other_options = {"split": "test"}
        X_columns = ["title", "description"]
        labels = ["label"]
        data = tfds.load("ag_news_subset", **other_options)
        config = FedDatasetConfig(
            seed=0,
            n_nodes=2,
            replacement=False,
            node_ids=["client_0", "client_1"],
        )
        # With batch_size
        other_options["batch_size"] = 20
        data = tfds.load("ag_news_subset", **other_options)
        build_data_and_check(data, config, X_columns, labels)

    def test_from_tfds_text_dataset_without_batchsize(self):
        import tensorflow_datasets as tfds

        def build_data_and_check(data, config, X_columns, labels):
            flex_dataset = FedDataDistribution.from_config_with_tfds_text_dataset(
                data, config, X_columns, labels
            )
            assert len(flex_dataset) == config.n_nodes
            assert len(flex_dataset["client_0"]) == len(flex_dataset["client_1"])

        other_options = {"split": "test"}
        X_columns = ["title", "description"]
        labels = ["label"]
        data = tfds.load("ag_news_subset", **other_options)
        config = FedDatasetConfig(
            seed=0,
            n_nodes=2,
            replacement=False,
            node_ids=["client_0", "client_1"],
        )
        # Without batch_size
        data = tfds.load("ag_news_subset", **other_options)
        build_data_and_check(data, config, X_columns, labels)

    def test_from_torchvision_dataset(self):
        from torchvision.datasets import MNIST

        data = MNIST(root="./torch_datasets", train=False, download=True)
        config = FedDatasetConfig(
            seed=0,
            n_nodes=2,
            replacement=False,
            node_ids=["client_0", "client_1"],
        )
        flex_dataset = FedDataDistribution.from_config_with_torchvision_dataset(
            data, config
        )
        assert len(flex_dataset) == config.n_nodes
        assert len(flex_dataset["client_0"]) == len(flex_dataset["client_1"])

    @pytest.mark.skipif(
        condition=os.getenv("GITHUB_ACTIONS") == "true",
        reason="Food101 is very huge and exceeds the RAM limit on GitHub.",
    )
    def test_from_torchvision_dataset_lazyly(self):
        from torchvision.datasets import Food101

        data = Food101(root="./torch_datasets", split="test", download=True)
        config = FedDatasetConfig(
            seed=0,
            n_nodes=2,
            replacement=False,
            node_ids=["client_0", "client_1"],
            labels_per_node=[[2, 3], [2]],
        )
        flex_dataset = FedDataDistribution.from_config_with_torchvision_dataset(
            data, config
        )
        assert len(flex_dataset) == config.n_nodes
        assert set(flex_dataset["client_0"].y_data) == {2, 3}
        assert set(flex_dataset["client_1"].y_data) == {2}
        # Non-Empty intersection
        assert (
            set(flex_dataset["client_1"].X_data._iterable_indexes)
            & set(flex_dataset["client_0"].X_data._iterable_indexes)
            == set()
        )

    def test_from_huggingface_text_dataset(self):
        from datasets import load_dataset

        data = load_dataset("ag_news", split="test")
        X_columns = ["text"]
        label_columns = ["label"]
        config = FedDatasetConfig(
            seed=0,
            replacement=False,
            node_ids=["client_0", "client_1"],
        )
        flex_dataset = FedDataDistribution.from_config_with_huggingface_dataset(
            data, config, X_columns, label_columns
        )
        assert len(flex_dataset) == config.n_nodes
        assert len(flex_dataset["client_0"]) == len(flex_dataset["client_1"])
        assert (
            len(flex_dataset["client_0"]) + len(flex_dataset["client_1"])
            == data.num_rows
        )
        # Empty intersection
        assert (
            set(flex_dataset["client_1"].X_data._iterable_indexes)
            & set(flex_dataset["client_0"].X_data._iterable_indexes)
            == set()
        )

    def test_from_huggingface_text_dataset_lazy(self):
        from datasets import load_dataset

        data = load_dataset("ag_news", split="test")
        X_columns = ["text"]
        label_columns = ["label"]
        config = FedDatasetConfig(
            seed=0,
            replacement=False,
            node_ids=["client_0", "client_1"],
        )
        flex_dataset = FedDataDistribution.from_config_with_huggingface_dataset(
            data, config, X_columns, label_columns
        )
        assert len(flex_dataset) == config.n_nodes
        assert len(flex_dataset["client_0"]) == len(flex_dataset["client_1"])
        assert (
            len(flex_dataset["client_0"]) + len(flex_dataset["client_1"])
            == data.num_rows
        )
        # Empty intersection
        assert (
            set(flex_dataset["client_1"].X_data._iterable_indexes)
            & set(flex_dataset["client_0"].X_data._iterable_indexes)
            == set()
        )

    def test_from_huggingface_text_dataset_lazy_str_no_subset(self):
        data = "ag_news;train"
        X_columns = ["text"]
        label_columns = ["label"]
        config = FedDatasetConfig(
            seed=0,
            replacement=False,
            node_ids=["client_0", "client_1"],
        )
        flex_dataset = FedDataDistribution.from_config_with_huggingface_dataset(
            data, config, X_columns, label_columns
        )
        assert len(flex_dataset) == config.n_nodes
        assert len(flex_dataset["client_0"]) == len(flex_dataset["client_1"])
        # Empty intersection
        assert (
            set(flex_dataset["client_1"].X_data._iterable_indexes)
            & set(flex_dataset["client_0"].X_data._iterable_indexes)
            == set()
        )

    def test_from_huggingface_text_dataset_lazy_str_and_subset(self):
        data = "tweet_eval;emoji;train"
        X_columns = ["text"]
        label_columns = ["label"]
        config = FedDatasetConfig(
            seed=0,
            replacement=False,
            node_ids=["client_0", "client_1"],
        )
        flex_dataset = FedDataDistribution.from_config_with_huggingface_dataset(
            data, config, X_columns, label_columns
        )
        assert len(flex_dataset) == config.n_nodes
        assert len(flex_dataset["client_0"]) == len(flex_dataset["client_1"])
        # Empty intersection
        assert (
            set(flex_dataset["client_1"].X_data._iterable_indexes)
            & set(flex_dataset["client_0"].X_data._iterable_indexes)
            == set()
        )

    def test_loading_fedmnist_digits_using_from_config(self):
        fed_data, test_data = load("federated_emnist", return_test=True, split="digits")
        assert isinstance(fed_data, FedDataset)
        assert isinstance(test_data, Dataset)
        num_samples = [len(fed_data[i]) for i in fed_data]
        total_samples = np.sum(num_samples)
        std = np.std(num_samples)
        mean = np.mean(num_samples)
        users = len(fed_data)
        assert users == 3579
        assert total_samples == 240000
        assert isclose(mean, 67.05, abs_tol=1e-1)
        assert isclose(std, 11.17, abs_tol=1e-1)

    def test_loading_fedmnist_letters_using_from_config(self):
        fed_data, test_data = load(
            "federated_emnist", return_test=True, split="letters"
        )
        assert isinstance(fed_data, FedDataset)
        assert isinstance(test_data, Dataset)
        num_samples = [len(fed_data[i]) for i in fed_data]
        total_samples = np.sum(num_samples)
        std = np.std(num_samples)
        mean = np.mean(num_samples)
        users = len(fed_data)
        assert users == 3585
        assert total_samples == 124800
        assert isclose(mean, 34.81, abs_tol=1e-1)
        assert isclose(std, 21.85, abs_tol=1e-1)

    @pytest.mark.skip(
        reason="CelebA dataset from torchvision has a limited amount of downloads per day allowed"
    )
    def test_loading_fedceleba_using_from_config(self):
        fed_data, test_data = load("federated_celeba", return_test=True)
        assert isinstance(fed_data, FedDataset)
        assert isinstance(test_data, Dataset)
        num_samples = [len(fed_data[i]) for i in fed_data]
        total_samples = np.sum(num_samples)
        std = np.std(num_samples)
        mean = np.mean(num_samples)
        users = len(fed_data)
        assert users == 8192
        assert total_samples == 162770
        assert isclose(mean, 19.87, abs_tol=1e-1)
        assert isclose(std, 8.92, abs_tol=1e-1)

    @pytest.mark.skipif(
        condition=os.getenv("GITHUB_ACTIONS") == "true",
        reason="Sentiment140 is very huge and exceeds the RAM limit on GitHub.",
    )
    def test_loading_fedsentiment_using_from_config(self):
        fed_data, test_data = load("federated_sentiment140", return_test=True)
        assert isinstance(fed_data, FedDataset)
        assert isinstance(test_data, Dataset)
        num_samples = [len(fed_data[i]) for i in fed_data]
        total_samples = np.sum(num_samples)
        std = np.std(num_samples)
        mean = np.mean(num_samples)
        users = len(fed_data)
        assert users == 659775
        assert total_samples == 1600000
        assert isclose(mean, 2.42, abs_tol=1e-1)
        assert isclose(std, 4.71, abs_tol=1e-1)

    @pytest.mark.skipif(
        condition=os.getenv("GITHUB_ACTIONS") == "true",
        reason="Sentiment140 is very huge and exceeds the RAM limit on GitHub.",
    )
    def test_loading_fedsentiment_using_from_config_lazyly(self):
        fed_data, test_data = load("federated_sentiment140", return_test=True)
        assert isinstance(fed_data, FedDataset)
        assert isinstance(test_data, Dataset)
        num_samples = [len(fed_data[i]) for i in fed_data]
        total_samples = np.sum(num_samples)
        std = np.std(num_samples)
        mean = np.mean(num_samples)
        users = len(fed_data)
        assert users == 659775
        assert total_samples == 1600000
        assert isclose(mean, 2.42, abs_tol=1e-1)
        assert isclose(std, 4.71, abs_tol=1e-1)

    def test_loading_fedshakespeare_using_from_config(self):
        fed_data, test_data = load("federated_shakespeare", return_test=True)
        assert isinstance(fed_data, FedDataset)
        assert isinstance(test_data, Dataset)
        num_samples = [len(fed_data[i]) for i in fed_data]
        total_samples = np.sum(num_samples)
        std = np.std(num_samples)
        mean = np.mean(num_samples)
        users = len(fed_data)
        assert users == 660
        assert total_samples == 3678451
        assert isclose(mean, 5573.41, abs_tol=1e-1)
        assert isclose(std, 6460.77, abs_tol=1e-1)

    def test_emnist_wrong_split_error(self):
        with pytest.raises(ValueError):
            load("emnist", split="weird")



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np

from flex.common import utils
from flex.data import Dataset


def emnist(out_dir: str = ".", split="digits", include_authors=False):
    from scipy.io import loadmat

    if split == "digits":
        url, filename, md5 = (
            utils.EMNIST_DIGITS_URL,
            utils.EMNIST_DIGITS_FILE,
            utils.EMNIST_DIGITS_MD5,
        )
    elif split == "letters":
        url, filename, md5 = (
            utils.EMNIST_LETTERS_URL,
            utils.EMNIST_LETTERS_FILE,
            utils.EMNIST_LETTERS_MD5,
        )
    else:
        raise ValueError(
            f"Unknown split: {split}. Available splits are 'digits' and 'letters'."
        )
    mnist_files = utils.download_dataset(
        url, filename, md5, out_dir=out_dir, extract=False, output=True
    )
    dataset = loadmat(mnist_files)["dataset"]
    train_writers = dataset["train"][0, 0]["writers"][0, 0]
    train_data = np.reshape(
        dataset["train"][0, 0]["images"][0, 0], (-1, 28, 28), order="F"
    )
    train_labels = np.squeeze(dataset["train"][0, 0]["labels"][0, 0])
    if include_authors:
        train_labels = np.asarray(
            [(label, train_writers[i][0]) for i, label in enumerate(train_labels)],
            dtype=np.int64,
        )

    test_writers = dataset["test"][0, 0]["writers"][0, 0]
    test_data = np.reshape(
        dataset["test"][0, 0]["images"][0, 0], (-1, 28, 28), order="F"
    )
    test_labels = np.squeeze(dataset["test"][0, 0]["labels"][0, 0])
    if include_authors:
        test_labels = np.asarray(
            [(label, test_writers[i][0]) for i, label in enumerate(test_labels)],
            dtype=np.int64,
        )
    train_data_object = Dataset.from_array(train_data, train_labels)
    test_data_object = Dataset.from_array(test_data, test_labels)
    return train_data_object, test_data_object


def shakespeare(out_dir: str = ".", include_actors=False):
    import json

    shakespeare_files = utils.download_dataset(
        utils.SHAKESPEARE_URL,
        utils.SHAKESPEARE_FILE,
        utils.SHAKESPEARE_MD5,
        out_dir=out_dir,
        extract=True,
        output=True,
    )
    train_files = filter(
        lambda n: "train" in n and n.endswith(".json"), shakespeare_files
    )
    train_x = []
    train_y = []
    for f in train_files:
        with open(f) as json_file:
            train_data = json.load(json_file)
        for user_id in train_data["users"]:
            node_ds = train_data["user_data"][user_id]
            if include_actors:
                train_y += [(y, user_id) for y in node_ds["y"]]
            else:
                train_y += node_ds["y"]
            train_x += node_ds["x"]
    test_files = filter(
        lambda n: "test" in n and n.endswith(".json"), shakespeare_files
    )
    test_x = []
    test_y = []
    for f in test_files:
        with open(f) as json_file:
            test_data = json.load(json_file)
        for user_id in test_data["users"]:
            node_ds = test_data["user_data"][user_id]
            if include_actors:
                test_y += [(y, user_id) for y in node_ds["y"]]
            else:
                test_y += node_ds["y"]
            test_x += node_ds["x"]

    return Dataset.from_array(train_x, train_y), Dataset.from_array(test_x, test_y)



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from copy import deepcopy

import numpy as np

from flex.data.dataset import Dataset
from flex.data.lazy_indexable import LazyIndexable


def normalize(node_dataset, *args, **kwargs):
    """Function that normalizes federated data.

    Args:
    -----
        node_dataset (Dataset): node_dataset  to normalize the data.

    Returns:
    --------
        Dataset: Returns the node_dataset with the X_data property normalized.
    """
    X_data = node_dataset.X_data.to_numpy()
    norms = np.linalg.norm(X_data, axis=0)
    norms = np.where(norms == 0, np.finfo(X_data.dtype).eps, norms)
    new_X_data = X_data / norms
    return Dataset.from_array(new_X_data, node_dataset.y_data.to_numpy())


def one_hot_encoding(node_dataset, *args, **kwargs):
    """Function that apply one hot encoding to the labels of a node_dataset.

    Args:
    -----
        node_dataset (Dataset): node_dataset to which apply one hot encode to her labels.

    Raises:
    -------
        ValueError: Raises value error if n_labels is not given in the kwargs argument.

    Returns:
    --------
        Dataset: Returns the node_dataset with the y_data property updated.
    """
    if "n_labels" not in kwargs:
        raise ValueError(
            "No number of labels given. The parameter n_labels must be given through kwargs."
        )
    y_data = node_dataset.y_data.to_numpy()
    n_labels = int(kwargs["n_labels"])
    one_hot_labels = np.zeros((y_data.size, n_labels))
    one_hot_labels[np.arange(y_data.size), y_data] = 1
    new__y_data = one_hot_labels
    return Dataset(
        X_data=deepcopy(node_dataset.X_data),
        y_data=LazyIndexable(new__y_data, len(new__y_data)),
    )



"""
Copyright (C) 2024  Instituto Andaluz Interuniversitario en Ciencia de Datos e Inteligencia Computacional (DaSCI).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import unittest

import pytest

from flex.model import FlexModel


class TestFlexModel(unittest.TestCase):
    def test_getter_actor_id(self):
        flex_model = FlexModel()
        flex_model.actor_id = "server"
        assert flex_model.actor_id == "server"

    def test_setter_actor_id(self):
        flex_model = FlexModel()
        flex_model.actor_id = "server"
        with pytest.raises(PermissionError):
            flex_model.actor_id = "client_1"

    def test_deleter_actor_id(self):
        flex_model = FlexModel()
        flex_model.actor_id = "server"
        del flex_model.actor_id
        assert flex_model.actor_id == "server"
